from typing import TypeVar

from jose import JWSError, JWTError
from lilya._internal._connection import Connection
from lilya.types import ASGIApp

from esmerald.config.jwt import JWTConfig
from esmerald.exceptions import AuthenticationError, NotAuthorized
from esmerald.middleware.authentication import AuthResult, BaseAuthMiddleware
from esmerald.security.jwt.token import Token

T = TypeVar("T")


class CommonJWTAuthMiddleware(BaseAuthMiddleware):  # pragma: no cover
    """
    The simple JWT authentication Middleware.
    """

    def __init__(
        self,
        app: ASGIApp,
        config: "JWTConfig",
        user_model: T,
    ):
        super().__init__(app)
        """
        The user is simply the class type to be queried from the Tortoise ORM.

        Example how to use:

            1. User table

                from esmerald.contrib.auth.tortoise.base_user import User as BaseUser

                class User(BaseUser):
                    ...

            2. Middleware

                from esmerald.contrib.auth.tortoise.middleware import JWTAuthMiddleware
                from esmerald.config import JWTConfig

                jwt_config = JWTConfig(...)

                class CustomJWTMidleware(JWTAuthMiddleware):
                    def __init__(self, app: "ASGIApp"):
                        super().__init__(app, config=jwt_config, user=User)

            3. The application
                from esmerald import Esmerald
                from myapp.middleware import CustomJWTMidleware

                app = Esmerald(routes=[...], middleware=[CustomJWTMidleware])

        """
        self.app = app
        self.config = config
        self.user_model = user_model

    async def authenticate(self, request: Connection) -> AuthResult:
        """
        Retrieves the header default of the config and validates against the decoding.

        Raises Authentication error if invalid.
        """
        token = request.headers.get(self.config.authorization_header, None)

        if not token or token is None:
            raise NotAuthorized(detail="Token not found in the request header")

        token_partition = token.partition(" ")
        token_type = token_partition[0]
        auth_token = token_partition[-1]

        if token_type not in self.config.auth_header_types:
            raise NotAuthorized(detail=f"'{token_type}' is not an authorized header.")

        try:
            token = Token.decode(
                token=auth_token,
                key=self.config.signing_key,
                algorithms=[self.config.algorithm],
            )
        except (JWSError, JWTError) as e:
            raise AuthenticationError(str(e)) from e

        user = await self.retrieve_user(token.sub)
        if not user:
            raise AuthenticationError("User not found.")
        return AuthResult(user=user)



from typing import Any, Dict, Literal, Optional, Union

from esmerald.openapi.enums import SecuritySchemeType
from esmerald.openapi.models import OAuthFlows
from esmerald.openapi.security.base import HTTPBase


class OAuth2(HTTPBase):
    """
    The OAuth2 scheme.

    For every parameter of the OAuthFlows, expects a OAuthFlow object type.

    Example:
        implicit: Optional[OAuthFlow] = OAuthFlow()
        password: Optional[OAuthFlow] = OAuthFlow()
        clientCredentials: Optional[OAuthFlow] = OAuthFlow()
        authorizationCode: Optional[OAuthFlow] = OAuthFlow()

        flows: OAuthFlows(
            implicit=implicit,
            password=password,
            clientCredentials=clientCredentials,
            authorizationCode=authorizationCode,
        )
    """

    def __init__(
        self,
        *,
        type_: Literal[
            "apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"
        ] = SecuritySchemeType.oauth2.value,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        name: Optional[str] = None,
        flows: Union[OAuthFlows, Dict[str, Dict[str, Any]]] = OAuthFlows(),
        **kwargs: Any,
    ):
        extra: Dict[Any, Any] = {}
        extra["flows"] = flows
        extra.update(kwargs)
        super().__init__(
            type_=type_,
            description=description,
            name=name,
            scheme_name=scheme_name or self.__class__.__name__,
            **extra,
        )



from typing import Any, Literal, Optional, Union

from pydantic import AnyUrl, ConfigDict

from esmerald.openapi.models import SecurityScheme


class HTTPBase(SecurityScheme):
    """
    Base for all HTTP security headers.
    """

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    def __init__(
        self,
        *,
        type_: Optional[Literal["apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"]] = None,
        bearerFormat: Optional[str] = None,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        in_: Optional[Literal["query", "header", "cookie"]] = None,
        name: Optional[str] = None,
        scheme: Optional[str] = None,
        openIdConnectUrl: Optional[Union[AnyUrl, str]] = None,
        **kwargs: Any,
    ):
        super().__init__(  # type: ignore
            type=type_,
            bearerFormat=bearerFormat,
            description=description,
            name=name,
            security_scheme_in=in_,
            scheme_name=scheme_name,
            scheme=scheme,
            openIdConnectUrl=openIdConnectUrl,
            **kwargs,
        )



import traceback
from contextlib import AsyncExitStack
from typing import Optional

from lilya.types import ASGIApp, Receive, Scope, Send

from esmerald.config import AsyncExitConfig
from esmerald.protocols.middleware import MiddlewareProtocol


class AsyncExitStackMiddleware(MiddlewareProtocol):
    def __init__(
        self,
        app: "ASGIApp",
        config: "AsyncExitConfig",
        debug: bool = False,
    ):
        """AsyncExitStack Middleware class.

        Args:
            app: The 'next' ASGI app to call.
            config: The AsyncExitConfig instance.
            debug: If the application should print the stack trace on any error.
        """
        super().__init__(app)
        self.app = app
        self.config = config
        self.debug = debug

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        if not AsyncExitStack:
            await self.app(scope, receive, send)  # pragma: no cover

        exception: Optional[Exception] = None
        async with AsyncExitStack() as stack:
            scope[self.config.context_name] = stack
            try:
                await self.app(scope, receive, send)
            except Exception as e:
                exception = e

        if exception and self.debug:
            traceback.print_exception(exception, exception, exception.__traceback__)  # type: ignore

        if exception:
            raise exception



from typing import List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from .operation import Operation
from .parameter import Parameter
from .reference import Reference
from .server import Server


class PathItem(BaseModel):
    """Describes the operations available on a single path.

    A Path Item MAY be empty, due to [ACL constraints](https://spec.openapis.org/oas/v3.1.0#securityFiltering).
    The path itself is still exposed to the documentation viewer
    but they will not know which operations and parameters are available.
    """

    ref: Optional[str] = Field(default=None, alias="$ref")
    """
    Allows for an external definition of this path item.
    The referenced structure MUST be in the format of a [Path Item Object](https://spec.openapis.org/oas/v3.1.0#pathItemObject).

    In case a Path Item Object field appears both in the defined object and the referenced object,
    the behavior is undefined.
    See the rules for resolving [Relative References](https://spec.openapis.org/oas/v3.1.0#relativeReferencesURI).
    """

    summary: Optional[str] = None
    """
    An optional, string summary, intended to apply to all operations in this path.
    """

    description: Optional[str] = None
    """
    An optional, string description, intended to apply to all operations in this path.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text representation.
    """

    get: Optional[Operation] = None
    """
    A definition of a GET operation on this path.
    """

    put: Optional[Operation] = None
    """
    A definition of a PUT operation on this path.
    """

    post: Optional[Operation] = None
    """
    A definition of a POST operation on this path.
    """

    delete: Optional[Operation] = None
    """
    A definition of a DELETE operation on this path.
    """

    options: Optional[Operation] = None
    """
    A definition of a OPTIONS operation on this path.
    """

    head: Optional[Operation] = None
    """
    A definition of a HEAD operation on this path.
    """

    patch: Optional[Operation] = None
    """
    A definition of a PATCH operation on this path.
    """

    trace: Optional[Operation] = None
    """
    A definition of a TRACE operation on this path.
    """

    servers: Optional[List[Server]] = None
    """
    An alternative `server` array to service all operations in this path.
    """

    parameters: Optional[List[Union[Parameter, Reference]]] = None
    """
    A list of parameters that are applicable for all the operations described under this path.
    These parameters can be overridden at the operation level, but cannot be removed there.
    The list MUST NOT include duplicated parameters.
    A unique parameter is defined by a combination of a [name](https://spec.openapis.org/oas/v3.1.0#parameterName) and [location](https://spec.openapis.org/oas/v3.1.0#parameterIn).
    The list can use the [Reference Object](https://spec.openapis.org/oas/v3.1.0#referenceObject) to link to parameters that are defined at the
    [OpenAPI Object's components/parameters](https://spec.openapis.org/oas/v3.1.0#componentsParameters).
    """
    model_config = ConfigDict(
        extra="ignore",
        populate_by_name=True,
        json_schema_extra={
            "examples": [
                {
                    "get": {
                        "description": "Returns pets based on ID",
                        "summary": "Find pets by ID",
                        "operationId": "getPetsById",
                        "responses": {
                            "200": {
                                "description": "pet response",
                                "content": {
                                    "*/*": {
                                        "schema": {
                                            "type": "array",
                                            "items": {"$ref": "#/components/schemas/Pet"},
                                        }
                                    }
                                },
                            },
                            "default": {
                                "description": "error payload",
                                "content": {
                                    "text/html": {
                                        "schema": {"$ref": "#/components/schemas/ErrorModel"}
                                    }
                                },
                            },
                        },
                    },
                    "parameters": [
                        {
                            "name": "id",
                            "in": "path",
                            "description": "ID of pet to use",
                            "required": True,
                            "schema": {"type": "array", "items": {"type": "string"}},
                            "style": "simple",
                        }
                    ],
                }
            ]
        },
    )



from typing import Any, Union

from esmerald.datastructures.json import JSON

try:
    from esmerald.datastructures.encoders import UJSON, OrJSON
except ImportError:
    UJSON = Any  # type: ignore[misc,assignment]
    OrJSON = Any  # type: ignore[misc,assignment]

EncoderType = Union[JSON, UJSON, OrJSON]



from typing import Any, Callable, Type, TypeVar

from pydantic_core import PydanticUndefined

T = TypeVar("T")


class Void:
    """A placeholder class."""


VoidType = Type[Void]
AnyCallable = Callable[..., Any]
Undefined = PydanticUndefined



from inspect import Signature
from typing import Any, Optional, Union

from pydantic import BaseModel, ConfigDict

from esmerald.enums import MediaType


class InternalResponse(BaseModel):
    """
    Response generated for non common return types.
    """

    media_type: Optional[Union[str, MediaType]] = None
    return_annotation: Optional[Any] = None
    signature: Optional[Signature] = None
    description: Optional[str] = None
    encoding: Optional[str] = None

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(annotation={self.media_type}, default={self.encoding})"



from typing import TYPE_CHECKING, Callable, Optional

from esmerald.exceptions import PermissionDenied
from esmerald.utils.helpers import is_async_callable

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.permissions import BasePermission
    from esmerald.requests import Request
    from esmerald.types import APIGateHandler


async def continue_or_raise_permission_exception(
    request: "Request",
    apiview: "APIGateHandler",
    permission: "BasePermission",
) -> None:
    """
    Check if the request has permission to access the API view.
    If not permitted, raise a PermissionDenied exception.
    """
    has_permission: Callable = permission.has_permission

    if not is_async_callable(has_permission):
        if not has_permission(request=request, apiview=apiview):
            permission_denied(
                request,
                message=getattr(permission, "message", None),
            )
    else:
        is_permission = await has_permission(request=request, apiview=apiview)
        if not is_permission:
            permission_denied(
                request,
                message=getattr(permission, "message", None),
            )


def permission_denied(request: "Request", message: Optional[str] = None) -> None:
    """
    If request is not permitted, determine what kind of exception to raise.
    """
    raise PermissionDenied(detail=message, status_code=403)



from typing import TYPE_CHECKING, Any, cast

from lilya._internal._connection import Connection as Connection  # noqa: F401
from lilya.datastructures import URL  # noqa
from lilya.requests import (
    ClientDisconnect as ClientDisconnect,  # noqa
    Request as LilyaRequest,  # noqa: F401
    empty_receive,
    empty_send,
)
from lilya.types import Receive, Scope, Send
from orjson import loads

from esmerald.typing import Void

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.applications import Esmerald
    from esmerald.conf.global_settings import EsmeraldAPISettings
    from esmerald.types import HTTPMethod


class Request(LilyaRequest):
    def __init__(
        self,
        scope: "Scope",
        receive: "Receive" = empty_receive,
        send: "Send" = empty_send,
    ):
        super().__init__(scope, receive, send)
        self._json: Any = Void

    @property
    def app(self) -> "Esmerald":
        return cast("Esmerald", self.scope["app"])

    @property
    def method(self) -> "HTTPMethod":
        return cast("HTTPMethod", self.scope["method"])

    @property
    def global_settings(self) -> Any:
        """
        Access to the global settings via `request.global_settings`.
        """
        assert (
            "global_settings" in self.scope
        ), "RequestSettingsMiddleware must be added to the middlewares"
        return cast("EsmeraldAPISettings", self.scope["global_settings"])

    @property
    def app_settings(self) -> Any:
        """
        Access to the app settings via `request.app_settings`.
        """
        assert (
            "app_settings" in self.scope
        ), "RequestSettingsMiddleware must be added to the middlewares"
        return cast("EsmeraldAPISettings", self.scope["app_settings"])

    async def json(self) -> Any:
        if self._json is Void:
            if "_body" in self.scope:
                body = self.scope["_body"]
            else:
                body = self.scope["_body"] = await self.body() or b"null"
            self._json = loads(body)
        return self._json

    def path_for(self, __name: str, **path_params: Any) -> Any:
        url: URL = super().path_for(__name, **path_params)
        return str(url)



import http.client
import inspect
import json
import warnings
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    _GenericAlias,
    cast,
)

from lilya._internal._path import clean_path
from lilya.middleware import DefineMiddleware
from lilya.routing import BasePath
from lilya.status import HTTP_422_UNPROCESSABLE_ENTITY
from lilya.transformers import TRANSFORMER_TYPES
from orjson import loads
from pydantic import AnyUrl
from pydantic.fields import FieldInfo
from pydantic.json_schema import GenerateJsonSchema, JsonSchemaValue
from typing_extensions import Literal

from esmerald.enums import MediaType
from esmerald.openapi.constants import METHODS_WITH_BODY, REF_PREFIX, REF_TEMPLATE
from esmerald.openapi.models import (
    Contact,
    Info,
    License,
    OpenAPI,
    Operation,
    Parameter,
    SecurityScheme,
)
from esmerald.openapi.responses import create_internal_response
from esmerald.openapi.utils import (
    STATUS_CODE_RANGES,
    VALIDATION_ERROR_DEFINITION,
    VALIDATION_ERROR_RESPONSE_DEFINITION,
    dict_update,
    get_definitions,
    get_schema_from_model_field,
    is_status_code_allowed,
)
from esmerald.params import Param
from esmerald.routing import gateways, router
from esmerald.routing._internal import convert_annotation_to_pydantic_model
from esmerald.typing import Undefined
from esmerald.utils.helpers import is_class_and_subclass, is_union

ADDITIONAL_TYPES = ["bool", "list", "dict"]
TRANSFORMER_TYPES_KEYS = list(TRANSFORMER_TYPES.keys())
TRANSFORMER_TYPES_KEYS += ADDITIONAL_TYPES


def get_flat_params(route: Union[router.HTTPHandler, Any]) -> List[Any]:
    """
    Gets all the neded params of the request and route.
    """
    path_params = [param.field_info for param in route.transformer.get_path_params()]
    cookie_params = [param.field_info for param in route.transformer.get_cookie_params()]
    header_params = [param.field_info for param in route.transformer.get_header_params()]

    handler_query_params = [
        param
        for param in route.transformer.get_query_params()
        if param.field_alias not in route.body_encoder_fields.keys()
    ]
    query_params = []
    for param in handler_query_params:
        is_union_or_optional = is_union(param.field_info.annotation)

        # Making sure all the optional and union types are included
        if is_union_or_optional:
            # field_info = should_skip_json_schema(param.field_info)
            query_params.append(param.field_info)

        else:
            if isinstance(param.field_info.annotation, _GenericAlias):
                query_params.append(param.field_info)
            elif (
                param.field_info.annotation.__class__.__name__ in TRANSFORMER_TYPES_KEYS
                or param.field_info.annotation.__name__ in TRANSFORMER_TYPES_KEYS
            ):
                query_params.append(param.field_info)

    return path_params + query_params + cookie_params + header_params


def get_openapi_security_schemes(schemes: Any) -> Tuple[dict, list]:
    """
    Builds the security schemas for OpenAPI.
    """
    security_definitions = {}
    operation_security = []

    for security_requirement in schemes:
        if inspect.isclass(security_requirement):
            security_requirement = security_requirement()

        if not isinstance(security_requirement, SecurityScheme):
            raise ValueError(
                "Security schemes must subclass from `esmerald.openapi.models.SecurityScheme`"
            )
        security_definition = security_requirement.model_dump(by_alias=True, exclude_none=True)
        security_name = security_requirement.scheme_name
        security_definitions[security_name] = security_definition
        operation_security.append({security_name: security_requirement})

    return security_definitions, operation_security


def get_fields_from_routes(
    routes: Sequence[BasePath], request_fields: Optional[List[FieldInfo]] = None
) -> List[FieldInfo]:
    """Extracts the fields from the given routes of Esmerald"""
    body_fields: List[FieldInfo] = []
    response_from_routes: List[FieldInfo] = []

    if not request_fields:
        request_fields = []

    for route in routes:
        if getattr(route, "include_in_schema", None) and isinstance(route, router.Include):
            request_fields.extend(get_fields_from_routes(route.routes, request_fields))
            continue

        if getattr(route, "include_in_schema", None) and isinstance(
            route, (gateways.Gateway, gateways.WebhookGateway)
        ):
            handler = cast(router.HTTPHandler, route.handler)

            if handler.data_field:
                body_fields.append(handler.data_field)

            if handler.response_models:
                for _, response in handler.response_models.items():
                    response_from_routes.append(response)

            # Get the params from the transformer
            params = get_flat_params(handler)
            if params:
                request_fields.extend(params)

    return list(body_fields + response_from_routes + request_fields)


def get_openapi_operation(
    *, route: gateways.Gateway, operation_ids: Set[str]
) -> Dict[str, Any]:  # pragma: no cover
    operation = Operation()
    operation.tags = route.handler.get_handler_tags()

    # Handle the routing summary
    if route.handler.summary:
        operation.summary = route.handler.summary
    else:
        operation.summary = route.handler.name.replace("_", " ").replace("-", " ").title()

    # Handle the handler description
    if route.handler.description:
        operation.description = route.handler.description
    else:
        operation.description = inspect.cleandoc(route.handler.fn.__doc__ or "")

    operation_id = getattr(route, "operation_id", None) or route.handler.operation_id

    if operation_id in operation_ids:
        message = (
            f"Duplicate Operation ID {operation_id} for function " + f"{route.handler.__name__}"
        )
        file_name = getattr(route.handler, "__globals__", {}).get("__file__")
        if file_name:
            message += f" at {file_name}"
        warnings.warn(message, stacklevel=1)
    operation_ids.add(operation_id)

    operation.operationId = operation_id
    if route.deprecated:
        operation.deprecated = route.deprecated
    elif route.handler.deprecated:
        operation.deprecated = route.handler.deprecated

    operation_schema = operation.model_dump(exclude_none=True, by_alias=True)
    return operation_schema


def get_openapi_operation_parameters(
    *,
    all_route_params: Sequence[FieldInfo],
    field_mapping: Dict[Tuple[FieldInfo, Literal["validation", "serialization"]], JsonSchemaValue],
) -> List[Dict[str, Any]]:  # pragma: no cover
    parameters = []
    for param in all_route_params:
        field_info = cast(Param, param)
        if not field_info.include_in_schema:
            continue

        param_schema = get_schema_from_model_field(
            field=param,
            field_mapping=field_mapping,
        )
        if field_info.default is not None and field_info.default is not Undefined:
            param_schema["default"] = field_info.default

        parameter = Parameter(  # type: ignore
            name=param.alias,
            param_in=field_info.in_.value,
            required=param.is_required(),
            schema=param_schema,  # type: ignore
        )
        if field_info.description:
            parameter.description = field_info.description
        if field_info.examples is not None:
            parameter.example = json.dumps(field_info.examples)
        if field_info.deprecated:
            parameter.deprecated = field_info.deprecated  # type: ignore
        parameters.append(parameter.model_dump(by_alias=True, exclude_none=True))
    return parameters


def get_openapi_operation_request_body(
    *,
    data_field: Optional[FieldInfo],
    field_mapping: Dict[Tuple[FieldInfo, Literal["validation", "serialization"]], JsonSchemaValue],
) -> Optional[Dict[str, Any]]:  # pragma: no cover
    if not data_field:
        return None

    assert isinstance(data_field, FieldInfo), "The 'data' needs to be a FieldInfo"
    schema = get_schema_from_model_field(field=data_field, field_mapping=field_mapping)

    field_info = data_field
    extra = cast("Dict[str, Any]", data_field.json_schema_extra)
    request_media_type = extra.get("media_type").value
    required = field_info.is_required()

    request_data_oai: Dict[str, Any] = {}
    if required:
        request_data_oai["required"] = required

    request_media_content: Dict[str, Any] = {"schema": schema}
    if field_info.examples is not None:
        request_media_content["example"] = json.dumps(field_info.examples)
    request_data_oai["content"] = {request_media_type: request_media_content}
    return request_data_oai


def get_openapi_path(
    *,
    route: Union[gateways.Gateway, gateways.WebhookGateway],
    operation_ids: Set[str],
    field_mapping: Dict[Tuple[FieldInfo, Literal["validation", "serialization"]], JsonSchemaValue],
    is_deprecated: bool = False,
) -> Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any]]:  # pragma: no cover
    path: Dict[str, Any] = {}
    security_schemes: Dict[str, Any] = {}
    definitions: Dict[str, Any] = {}

    assert route.handler.methods is not None, "Methods must be a list"
    route_response_media_type: str = None
    handler: router.HTTPHandler = cast("router.HTTPHandler", route.handler)

    if not handler.response_class:
        internal_response = create_internal_response(handler)
        route_response_media_type = internal_response.media_type
    else:
        assert (
            handler.response_class.media_type is not None
        ), "`media_type` is required in the response class."
        route_response_media_type = handler.response_class.media_type

    # If routes do not want to be included in the schema generation
    if not route.include_in_schema or not handler.include_in_schema:
        return path, security_schemes, definitions

    # For each method
    for method in route.handler.methods:
        operation = get_openapi_operation(route=route, operation_ids=operation_ids)  # type: ignore
        # If the parent if marked as deprecated, it takes precedence
        if is_deprecated or route.deprecated:
            operation["deprecated"] = is_deprecated if is_deprecated else route.deprecated

        parameters: List[Dict[str, Any]] = []
        security_definitions, operation_security = get_openapi_security_schemes(
            handler.get_security_schemes()
        )

        if operation_security:
            operation.setdefault("security", []).extend(operation_security)

        if security_definitions:
            security_schemes.update(security_definitions)

        all_route_params = get_flat_params(handler)
        operation_parameters = get_openapi_operation_parameters(
            all_route_params=all_route_params,
            field_mapping=field_mapping,
        )
        parameters.extend(operation_parameters)

        if parameters:
            all_parameters = {(param["in"], param["name"]): param for param in parameters}
            required_parameters = {
                (param["in"], param["name"]): param
                for param in parameters
                if param.get("required")
            }
            all_parameters.update(required_parameters)
            operation["parameters"] = list(all_parameters.values())

        if method in METHODS_WITH_BODY:
            request_data_oai = get_openapi_operation_request_body(
                data_field=handler.data_field,
                field_mapping=field_mapping,
            )
            if request_data_oai:
                operation["requestBody"] = request_data_oai

        status_code = str(handler.status_code)
        operation.setdefault("responses", {}).setdefault(status_code, {})[
            "description"
        ] = handler.response_description

        # Media type
        if route_response_media_type and is_status_code_allowed(handler.status_code):
            response_schema = (
                {"type": "string"} if handler.status_code not in handler.responses else {}
            )

            operation.setdefault("responses", {}).setdefault(status_code, {}).setdefault(
                "content", {}
            ).setdefault(route_response_media_type, {})["schema"] = response_schema

        # Additional responses
        if handler.response_models:
            operation_responses = operation.setdefault("responses", {})
            for additional_status_code, _ in handler.response_models.items():
                process_response = handler.responses[additional_status_code].model_copy()
                status_code_key = str(additional_status_code).upper()

                if status_code_key == "DEFAULT":
                    status_code_key = "default"

                openapi_response = operation_responses.setdefault(status_code_key, {})

                field = handler.response_models.get(additional_status_code)
                additional_field_schema: Optional[Dict[str, Any]] = None
                model_schema = process_response.model_json_schema()

                if field:
                    additional_field_schema = get_schema_from_model_field(
                        field=field, field_mapping=field_mapping
                    )
                    media_type = route_response_media_type or MediaType.JSON.value
                    additional_schema = (
                        model_schema.setdefault("content", {})
                        .setdefault(media_type, {})
                        .setdefault("schema", {})
                    )
                    dict_update(additional_schema, additional_field_schema)

                # status
                status_text = (
                    process_response.status_text
                    or STATUS_CODE_RANGES.get(str(additional_status_code).upper())
                    or http.client.responses.get(int(additional_status_code))
                )
                description = (
                    process_response.description
                    or openapi_response.get("description")
                    or status_text
                    or "Additional Response"
                )
                dict_update(openapi_response, model_schema)
                openapi_response["description"] = description

        # Convert to automatic response detection if none is provided by the
        # responses of the handler.
        if handler.handler_signature.return_annotation:
            response_schema = convert_annotation_to_pydantic_model(
                handler.handler_signature.return_annotation
            )

            if (
                hasattr(response_schema, "model_json_schema")
                and status_code not in handler.responses
                and int(status_code) not in handler.responses
            ):
                operation["responses"][status_code]["content"][route_response_media_type][
                    "schema"
                ] = response_schema.model_json_schema()

        http422 = str(HTTP_422_UNPROCESSABLE_ENTITY)
        if (all_route_params or handler.data_field) and not any(
            status in operation["responses"] for status in {http422, "4XX", "default"}
        ):
            operation["responses"][http422] = {
                "description": "Validation Error",
                "content": {
                    "application/json": {"schema": {"$ref": REF_PREFIX + "HTTPValidationError"}}
                },
            }
            if "ValidationError" not in definitions:
                definitions.update(
                    {
                        "ValidationError": VALIDATION_ERROR_DEFINITION,
                        "HTTPValidationError": VALIDATION_ERROR_RESPONSE_DEFINITION,
                    }
                )
        path[method.lower()] = operation
    return path, security_schemes, definitions


def should_include_in_schema(route: router.Include) -> bool:
    """
    Checks if a specifc object should be included in the schema
    """
    from esmerald import ChildEsmerald, Esmerald

    if not route.include_in_schema:
        return False

    if not is_middleware_app(route):
        return True

    if (
        isinstance(route.app, (Esmerald, ChildEsmerald))
        or (
            is_class_and_subclass(route.app, Esmerald)
            or is_class_and_subclass(route.app, ChildEsmerald)
        )
    ) and not getattr(route.app, "enable_openapi", False):
        return False
    if (
        isinstance(route.app, (Esmerald, ChildEsmerald))
        or (
            is_class_and_subclass(route.app, Esmerald)
            or is_class_and_subclass(route.app, ChildEsmerald)
        )
    ) and not getattr(route.app, "include_in_schema", False):
        return False

    return True


def is_middleware_app(route: router.Include) -> bool:
    """
    Checks if the app is a middleware or a router
    """
    from esmerald import MiddlewareProtocol

    return bool(isinstance(route.app, (DefineMiddleware, MiddlewareProtocol)))


def get_openapi(
    *,
    app: Any,
    title: str,
    version: str,
    openapi_version: str = "3.1.0",
    summary: Optional[str] = None,
    description: Optional[str] = None,
    routes: Sequence[BasePath],
    tags: Optional[List[str]] = None,
    servers: Optional[List[Dict[str, Union[str, Any]]]] = None,
    terms_of_service: Optional[Union[str, AnyUrl]] = None,
    contact: Optional[Contact] = None,
    license: Optional[License] = None,
    webhooks: Optional[Sequence[BasePath]] = None,
) -> Dict[str, Any]:  # pragma: no cover
    """
    Builds the whole OpenAPI route structure and object
    """
    from esmerald import ChildEsmerald, Esmerald

    info = Info(title=title, version=version)
    if summary:
        info.summary = summary
    if description:
        info.description = description
    if terms_of_service:
        info.termsOfService = terms_of_service  # type: ignore
    if contact:
        info.contact = contact
    if license:
        info.license = license

    output: Dict[str, Any] = {
        "openapi": openapi_version,
        "info": info.model_dump(exclude_none=True, by_alias=True),
    }

    if servers:
        output["servers"] = servers

    components: Dict[str, Dict[str, Any]] = {}
    paths: Dict[str, Dict[str, Any]] = {}
    webhooks_paths: Dict[str, Dict[str, Any]] = {}
    operation_ids: Set[str] = set()
    all_fields = get_fields_from_routes(list(routes or []) + list(webhooks or []))
    schema_generator = GenerateJsonSchema(ref_template=REF_TEMPLATE)
    field_mapping, definitions = get_definitions(
        fields=all_fields,
        schema_generator=schema_generator,
    )

    # Iterate through the routes
    def iterate_routes(
        app: Any,
        routes: Sequence[BasePath],
        definitions: Any = None,
        components: Any = None,
        prefix: Optional[str] = "",
        is_webhook: bool = False,
        is_deprecated: bool = False,
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        for route in routes:
            if app.router.deprecated:
                is_deprecated = True

            if isinstance(route, router.Include):
                if hasattr(route, "app"):
                    if not should_include_in_schema(route):
                        continue

                # For external middlewares
                if getattr(route.app, "routes", None) is None and not is_middleware_app(route):
                    continue

                if hasattr(route, "app") and isinstance(route.app, (Esmerald, ChildEsmerald)):
                    route_path = clean_path(prefix + route.path)

                    definitions, components = iterate_routes(
                        app,
                        route.app.routes,
                        definitions,
                        components,
                        prefix=route_path,
                        is_deprecated=is_deprecated if is_deprecated else route.deprecated,
                    )
                else:
                    route_path = clean_path(prefix + route.path)
                    definitions, components = iterate_routes(
                        app,
                        route.routes,
                        definitions,
                        components,
                        prefix=route_path,
                        is_deprecated=is_deprecated if is_deprecated else route.deprecated,
                    )
                continue

            if isinstance(route, (gateways.Gateway, gateways.WebhookGateway)):
                result = get_openapi_path(
                    route=route,
                    operation_ids=operation_ids,
                    field_mapping=field_mapping,
                    is_deprecated=is_deprecated,
                )
                if result:
                    path, security_schemes, path_definitions = result
                    if path:
                        if is_webhook:
                            webhooks_paths.setdefault(route.path, {}).update(path)
                        else:
                            route_path = clean_path(prefix + route.path_format)
                            paths.setdefault(route_path, {}).update(path)
                    if security_schemes:
                        components.setdefault("securitySchemes", {}).update(security_schemes)
                    if path_definitions:
                        definitions.update(path_definitions)

        return definitions, components

    definitions, components = iterate_routes(
        app=app, routes=routes, definitions=definitions, components=components
    )

    if webhooks:
        definitions, components = iterate_routes(
            app=app,
            routes=webhooks,
            definitions=definitions,
            components=components,
            is_webhook=True,
        )

    if definitions:
        components["schemas"] = {k: definitions[k] for k in sorted(definitions)}
    if components:
        output["components"] = components
    output["paths"] = paths
    if webhooks_paths:
        output["webhooks"] = webhooks_paths
    if tags:
        output["tags"] = tags

    openapi = OpenAPI(**output)
    model_dump = openapi.model_dump_json(by_alias=True, exclude_none=True)
    return cast(Dict[str, Any], loads(model_dump))



from typing import TYPE_CHECKING, Any, List, Union

from lilya.templating.jinja import Jinja2Template

from esmerald.exceptions import MissingDependency

if TYPE_CHECKING:  # pragma: no cover
    from pydantic import DirectoryPath

try:
    from jinja2 import Environment
except ImportError as exc:  # pragma: no cover
    raise MissingDependency("jinja2 is not installed") from exc


class JinjaTemplateEngine(Jinja2Template):
    def __init__(
        self,
        directory: Union["DirectoryPath", List["DirectoryPath"]],
        env: Union[Environment, None] = None,
        **env_options: Any,
    ) -> None:
        super().__init__(directory=directory, env=env, **env_options)



from datetime import datetime
from typing import Any, Type, cast

import mongoz

from esmerald.contrib.auth.hashers import check_password, is_password_usable, make_password


class AbstractUser(mongoz.Document):
    """
    Base model used for a custom user of any application.
    """

    first_name: str = mongoz.String(max_length=150)
    last_name: str = mongoz.String(max_length=150)
    username: str = mongoz.String(max_length=150, unique=True)
    email: str = mongoz.Email(max_length=120, unique=True)
    password: str = mongoz.String(max_length=128)
    last_login: datetime = mongoz.DateTime(null=True)
    is_active: bool = mongoz.Boolean(default=True)  # type: ignore
    is_staff: bool = mongoz.Boolean(default=False)  # type: ignore
    is_superuser: bool = mongoz.Boolean(default=False)  # type: ignore

    # Stores the raw password if set_password() is called so that it can
    # be passed to password_changed() after the model is saved.
    _password = None

    class Meta:
        abstract = True

    @property
    async def is_authenticated(self) -> bool:
        """
        Always return True.
        """
        return True

    async def set_password(self: "AbstractUser", raw_password: str) -> "mongoz.Document":
        self.password = make_password(raw_password)
        self._password = raw_password
        return await self.update(password=make_password(raw_password))

    async def check_password(self: "AbstractUser", raw_password: str) -> bool:
        """
        Return a boolean of whether the raw_password was correct. Handles
        hashing formats behind the scenes.
        """

        async def setter(raw_password: str) -> None:
            await self.set_password(raw_password)
            # Password hash upgrades shouldn't be considered password changes.
            self._password = None
            await self.update(password=self.password)

        return await check_password(raw_password, str(self.password), setter)

    async def set_unusable_password(self) -> None:
        # Set a value that will never be a valid hash
        self.password = make_password(None)

    async def has_usable_password(self) -> bool:
        """
        Return False if set_unusable_password() has been called for this user.
        """
        return is_password_usable(self.password)

    @classmethod
    async def _create_user(
        cls: Type["mongoz.Document"], username: str, email: str, password: str, **extra_fields: Any
    ) -> "AbstractUser":
        """
        Create and save a user with the given username, email, and password.
        """
        if not username:
            raise ValueError("The given username must be set")
        password = make_password(password)
        user: "mongoz.Document" = await cls.objects.create(
            username=username, email=email, password=password, **extra_fields
        )
        return cast("AbstractUser", user)

    @classmethod
    async def create_user(
        cls,
        username: str,
        email: str,
        password: str,
        **extra_fields: Any,
    ) -> "AbstractUser":
        extra_fields.setdefault("is_staff", False)
        extra_fields.setdefault("is_superuser", False)
        return await cls._create_user(username, email, password, **extra_fields)

    @classmethod
    async def create_superuser(
        cls,
        username: str,
        email: str,
        password: str,
        **extra_fields: Any,
    ) -> "AbstractUser":
        extra_fields.setdefault("is_staff", True)
        extra_fields.setdefault("is_superuser", True)

        if extra_fields.get("is_staff") is not True:
            raise ValueError("Superuser must have is_staff=True.")
        if extra_fields.get("is_superuser") is not True:
            raise ValueError("Superuser must have is_superuser=True.")

        return await cls._create_user(username, email, password, **extra_fields)


class User(AbstractUser):
    """
    Implementation of the Abstract user and can be used directly.
    """

    ...



import inspect
import os
import sys
from typing import TYPE_CHECKING, Any, Optional, Union

import click
from lilya._internal._path import clean_path
from rich.console import Console
from rich.table import Table

from esmerald import Gateway
from esmerald.core.directives.constants import ESMERALD_DISCOVER_APP
from esmerald.core.directives.env import DirectiveEnv
from esmerald.core.terminal import OutputColour, Print, Terminal
from esmerald.enums import HttpMethod
from esmerald.routing.apis.base import View

if TYPE_CHECKING:
    from lilya.routing import BasePath

    from esmerald.applications import ChildEsmerald, Esmerald
    from esmerald.routing.router import Router

printer = Print()
writer = Terminal()
console = Console()

DOCS_ELEMENTS = [
    "/swagger",
    "/redoc",
    "/openapi.json",
    "/openapi.yaml",
    "/openapi.yml",
    "/elements",
]


def get_http_verb(mapping: Any) -> str:
    if getattr(mapping, "get", None):
        return HttpMethod.GET.value
    elif getattr(mapping, "post", None):
        return HttpMethod.POST.value
    elif getattr(mapping, "put", None):
        return HttpMethod.PUT.value
    elif getattr(mapping, "patch", None):
        return HttpMethod.PATCH.value
    elif getattr(mapping, "delete", None):
        return HttpMethod.DELETE.value
    elif getattr(mapping, "header", None):
        return HttpMethod.HEAD.value
    return HttpMethod.GET.value


@click.command(name="show_urls")
def show_urls(env: DirectiveEnv) -> None:
    """Shows the information regarding the urls of a given application

    How to run: `esmerald show_urls`

    Example: `esmerald show_urls`
    """
    if os.getenv(ESMERALD_DISCOVER_APP) is None and getattr(env, "app", None) is None:
        error = (
            "You cannot specify a custom directive without specifying the --app or setting "
            "ESMERALD_DEFAULT_APP environment variable."
        )
        printer.write_error(error)
        sys.exit(1)

    app = env.app
    table = Table(title=app.app_name)
    table = get_routes_table(app, table)
    printer.write(table)


def get_routes_table(app: Optional[Union["Esmerald", "ChildEsmerald"]], table: Table) -> Table:
    """Prints the routing system"""
    table.add_column("Path", style=OutputColour.GREEN, vertical="middle")
    table.add_column("Path Parameters", style=OutputColour.BRIGHT_CYAN, vertical="middle")
    table.add_column("Name", style=OutputColour.CYAN, vertical="middle")
    table.add_column("Type", style=OutputColour.YELLOW, vertical="middle")
    table.add_column("HTTP Methods", style=OutputColour.RED, vertical="middle")

    def parse_routes(
        app: Optional[Union["Esmerald", "ChildEsmerald", "Router", "BasePath"]],
        table: Table,
        route: Optional[Any] = None,
        prefix: Optional[str] = "",
    ) -> None:
        if getattr(app, "routes", None) is None:
            return

        for route in app.routes:  # type: ignore
            if isinstance(route, Gateway):
                # Path
                path = clean_path(prefix + route.path)

                if any(element in path for element in DOCS_ELEMENTS):
                    continue

                # Type
                if not isinstance(route.handler, View):
                    if inspect.iscoroutinefunction(route.handler.fn):
                        fn_type = "async"
                    else:
                        fn_type = "sync"

                # Http methods
                http_methods = ", ".join(sorted(route.methods))
                parameters = ", ".join(sorted(route.stringify_parameters))
                table.add_row(path, parameters, route.name, fn_type, http_methods)
                continue

            route_app = getattr(route, "app", None)
            if not route_app:
                continue

            path = clean_path(prefix + route.path)  # type: ignore
            if any(element in path for element in DOCS_ELEMENTS):
                continue

            parse_routes(route, table, prefix=f"{path}")

    parse_routes(app, table)
    return table



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Iterator, Optional

from typing_extensions import Annotated, Doc

from esmerald.protocols.extension import ExtensionProtocol

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.applications import Esmerald


class Pluggable:
    """
    The `Pluggable` is used to create an `Esmerald` pluggable from an `Extension`.
    When Esmerald receives pluggables, it hooks them into the system and allows
    the access via anywhere in the application.

    Read more about the [Pluggables](https://esmerald.dev/pluggables/) and learn how to use them.

    **Example**

    ```python
    from typing import Optional

    from loguru import logger
    from pydantic import BaseModel

    from esmerald import Esmerald, Extension, Pluggable
    from esmerald.types import DictAny


    class PluggableConfig(BaseModel):
        name: str


    class MyExtension(Extension):
        def __init__(
            self, app: Optional["Esmerald"] = None, config: PluggableConfig = None, **kwargs: "DictAny"
        ):
            super().__init__(app, **kwargs)
            self.app = app

        def extend(self, config: PluggableConfig) -> None:
            logger.success(f"Successfully passed a config {config.name}")


    my_config = PluggableConfig(name="my extension")

    pluggable = Pluggable(MyExtension, config=my_config)

    app = Esmerald(routes=[], pluggables={"my-extension": pluggable})
    ```
    """

    def __init__(self, cls: "Extension", **options: Any):
        self.cls = cls
        self.options = options

    def __iter__(self) -> Iterator:
        iterator = (self.cls, self.options)
        return iter(iterator)

    def __repr__(self) -> str:  # pragma: no cover
        name = self.__class__.__name__
        options = [f"{key}={value!r}" for key, value in self.options.items()]
        args = ", ".join([self.__class__.__name__] + options)
        return f"{name}({args})"


class BaseExtension(ABC, ExtensionProtocol):  # pragma: no cover
    """
    The base for any Esmerald plugglable.
    """

    def __init__(
        self,
        app: Annotated[
            Optional["Esmerald"],
            Doc(
                """
                An `Esmerald` application instance or subclasses of Esmerald.
                """
            ),
        ] = None,
        **kwargs: Annotated[
            Any,
            Doc("""Any additional kwargs needed."""),
        ],
    ):
        super().__init__(app, **kwargs)
        self.app = app

    @abstractmethod
    def extend(self, **kwargs: "Any") -> None:
        raise NotImplementedError("plug must be implemented by the subclasses.")


class Extension(BaseExtension):
    """
    `Extension` object is the one being used to add the logic that will originate
    the `pluggable` in the application.

    The `Extension` **must implement** the `extend` function.

    Read more about the [Extension](https://esmerald.dev/pluggables/#extension) and learn how to use it.

    **Example**

    ```python
    from typing import Optional

    from esmerald import Esmerald, Extension
    from esmerald.types import DictAny


    class MyExtension(Extension):
        def __init__(self, app: Optional["Esmerald"] = None, **kwargs: "DictAny"):
            super().__init__(app, **kwargs)
            self.app = app
            self.kwargs = kwargs

        def extend(self, **kwargs: "DictAny") -> None:
            '''
            Function that should always be implemented when extending
            the Extension class or a `NotImplementedError` is raised.
            '''
            # Do something here
    ```
    """



from typing import TypeVar

from lilya.types import Receive, Scope, Send
from typing_extensions import Protocol, runtime_checkable

T = TypeVar("T")


@runtime_checkable
class InterceptorProtocol(Protocol):  # pragma: no cover
    """
    Generic object serving the base for interception of messages,
    before reaching the handler. This is inspired by the AOP (Aspect Oriented Programming).

    The interceptor is handled between the call and the API handler itself and acts on it.

    An interceptor could be anything from logging to rerouting or even input sanitizing.
    """

    async def intercept(self, scope: "Scope", receive: "Receive", send: "Send") -> None: ...



from typing import Optional

from pydantic import AnyUrl, BaseModel, ConfigDict


class License(BaseModel):
    """License information for the exposed API."""

    model_config = ConfigDict(
        extra="ignore",
        json_schema_extra={
            "examples": [
                {"name": "Apache 2.0", "identifier": "Apache-2.0"},
                {"name": "Apache 2.0", "url": "https://www.apache.org/licenses/LICENSE-2.0.html"},
            ]
        },
    )

    name: str
    """
    **REQUIRED**. The license name used for the API.
    """

    identifier: Optional[str] = None
    """
    An [SPDX](https://spdx.org/spdx-specification-21-web-version#h.jxpfx0ykyb60) license expression for the API.
    The `identifier` field is mutually exclusive of the `url` field.
    """

    url: Optional[AnyUrl] = None
    """
    A URL to the license used for the API.
    This MUST be in the form of a URL.
    The `url` field is mutually exclusive of the `identifier` field.
    """



from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class Reference(BaseModel):
    """A simple object to allow referencing other components in the OpenAPI
    document, internally and externally.

    The `$ref` string value contains a URI [RFC3986](https://tools.ietf.org/html/rfc3986),
    which identifies the location of the value being referenced.

    See the rules for resolving [Relative References](https://spec.openapis.org/oas/v3.1.0#relativeReferencesURI).
    """

    ref: str = Field(alias="$ref")
    """**REQUIRED**. The reference identifier. This MUST be in the form of a URI."""

    summary: Optional[str] = None
    """
    A short summary which by default SHOULD override that of the referenced component.
    If the referenced object-type does not allow a `summary` field, then this field has no effect.
    """

    description: Optional[str] = None
    """
    A description which by default SHOULD override that of the referenced component.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text representation.
    If the referenced object-type does not allow a `description` field, then this field has no effect.
    """
    model_config = ConfigDict(
        extra="ignore",
        populate_by_name=True,
        json_schema_extra={
            "examples": [
                {"$ref": "#/components/schemas/Pet"},
                {"$ref": "Pet.json"},
                {"$ref": "definitions.json#/Pet"},
            ]
        },
    )



from typing import TYPE_CHECKING, Union

from lilya.context import Context as LilyaContext
from typing_extensions import Annotated, Doc

if TYPE_CHECKING:
    from esmerald import Request
    from esmerald.routing.router import HTTPHandler, WebSocketHandler


class Context(LilyaContext):
    """
    `Context` class is used for the handlers context of the
    scope of the call.

    When a `context` is passed through the handler,
    the context will be aware of the decoractor handler
    itself.

    You will probably not need the context or change it but it is here in
    case you decide to use it.

    !!! Tip
        The context only exists in the handlers and nothing else which you can also
        see it as `request context` sort of approach.

    **Example**

    ```python
    from esmerald import Esmerald, Context, get


    @get()
    async def home(context: Context) -> None:
        ...
    ```
    """

    def __init__(
        self,
        __handler__: Annotated[
            Union["HTTPHandler", "WebSocketHandler"],
            Doc(
                """
                The [handler](https://esmerald.dev/routing/handlers/) where the context will be. placed.

                To avoid any adulteration of the the original route handler, the context performs
                a shallow copy of the original handler itself.
                """
            ),
        ],
        __request__: Annotated[
            "Request",
            Doc(
                """
                A [Request](https://esmerald.dev/references/request/) class object.
                """
            ),
        ],
    ) -> None:
        super().__init__(__handler__, __request__)



from datetime import datetime, timezone as dtimezone
from typing import Any, Callable, Dict, Type, Union

from asyncz.schedulers import AsyncIOScheduler
from asyncz.schedulers.types import SchedulerType
from asyncz.triggers.types import TriggerType
from asyncz.typing import Undefined, undefined

from esmerald.conf import settings
from esmerald.contrib.schedulers.base import SchedulerConfig
from esmerald.exceptions import ImproperlyConfigured
from esmerald.utils.module_loading import import_string

SchedulerCallable = Callable[..., Any]


class AsynczConfig(SchedulerConfig):
    """
    Implements an integration with Asyncz, allowing to
    customise the scheduler with the provided configurations.
    """

    def __init__(
        self,
        scheduler_class: Type[SchedulerType] = AsyncIOScheduler,
        tasks: Union[Dict[str, str]] = None,
        timezone: Union[dtimezone, str, None] = None,
        configurations: Union[Dict[str, Dict[str, str]], None] = None,
        **kwargs: Dict[str, Any],
    ):
        """
        Initializes the AsynczConfig object.

        Args:
            scheduler_class: The class of the scheduler to be used.
            tasks: A dictionary of tasks to be registered in the scheduler.
            timezone: The timezone to be used by the scheduler.
            configurations: Extra configurations to be passed to the scheduler.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)
        self.scheduler_class = scheduler_class
        self.tasks = tasks or {}
        self.timezone = timezone
        self.configurations = configurations
        self.options = kwargs

        for task, module in self.tasks.items():
            if not isinstance(task, str) or not isinstance(module, str):
                raise ImproperlyConfigured("The dict of tasks must be Dict[str, str].")

        # Load the scheduler object
        self.handler = self.get_scheduler(
            scheduler=self.scheduler_class,
            timezone=self.timezone,
            configurations=self.configurations,
            **self.options,
        )

        self.register_tasks(tasks=self.tasks)

    def register_tasks(self, tasks: Dict[str, str]) -> None:
        """
        Registers the tasks in the Scheduler.

        Args:
            tasks: A dictionary of tasks to be registered in the scheduler.
        """
        for task, _module in tasks.items():
            imported_task = f"{_module}.{task}"
            scheduled_task: "Task" = import_string(imported_task)

            if not scheduled_task.is_enabled:
                continue

            try:
                scheduled_task.add_task(self.handler)
            except Exception as e:
                raise ImproperlyConfigured(str(e)) from e

    def get_scheduler(
        self,
        scheduler: Type[SchedulerType],
        timezone: Union[dtimezone, str, None] = None,
        configurations: Union[Dict[str, Any], None] = None,
        **options: Dict[str, Any],
    ) -> SchedulerType:
        """
        Initiates the scheduler from the given time.
        If no value is provided, it will default to AsyncIOScheduler.

        The value of `scheduler_class` can be overwritten by any esmerald custom settings.

        Args:
            scheduler: The class of the scheduler to be used.
            timezone: The timezone instance.
            configurations: A dictionary with extra configurations to be passed to the scheduler.
            **options: Additional options.

        Returns:
            SchedulerType: An instance of a Scheduler.
        """
        if not timezone:
            timezone = settings.timezone

        if not configurations:
            return scheduler(timezone=timezone, **options)  # type: ignore

        return scheduler(global_config=configurations, timezone=timezone, **options)  # type: ignore

    async def start(self, **kwargs: Any) -> None:
        """
        Starts the scheduler.

        Args:
            **kwargs: Additional keyword arguments.
        """
        self.handler.start(**kwargs)

    async def shutdown(self, **kwargs: Any) -> None:
        """
        Shuts down the scheduler.

        Args:
            **kwargs: Additional keyword arguments.
        """
        self.handler.shutdown(**kwargs)


class Task:
    """
    Base for the scheduler decorator that will auto discover the
    tasks in the application and add them to the internal scheduler.
    """

    def __init__(
        self,
        fn: Any = None,
        *,
        name: Union[str, None] = None,
        trigger: Union[TriggerType, None] = None,
        id: Union[str, None] = None,
        mistrigger_grace_time: Union[int, Undefined, None] = undefined,
        coalesce: Union[bool, Undefined] = undefined,
        max_instances: Union[int, Undefined, None] = undefined,
        next_run_time: Union[datetime, str, Undefined, None] = undefined,
        store: Union[str, None] = None,
        executor: Union[str, None] = None,
        replace_existing: bool = False,
        args: Union[Any, None] = None,
        kwargs: Union[Dict[str, Any], None] = None,
        is_enabled: bool = True,
    ) -> None:
        """
        Initializes a new instance of the `Task` class for the  Scheduler.

        Args:
            name (str, optional): Textual description of the task.
            trigger (TriggerType, optional): An instance of a trigger class.
            id (str, optional): Explicit identifier for the task.
            mistrigger_grace_time (int, optional): Seconds after the designated runtime that the task is still allowed to be run
                (or None to allow the task to run no matter how late it is).
            coalesce (bool, optional): Run once instead of many times if the scheduler determines that the task should be run more than once in succession.
            max_instances (int, optional): Maximum number of concurrently running instances allowed for this task.
            next_run_time (datetime, optional): When to first run the task, regardless of the trigger (pass None to add the task as paused).
            store (str, optional): Alias of the task store to store the task in.
            executor (str, optional): Alias of the executor to run the task with.
            replace_existing (bool, optional): True to replace an existing task with the same id
                (but retain the number of runs from the existing one).
            args (Any, optional): List of positional arguments to call func with.
            kwargs (Dict[str, Any], optional): Dict of keyword arguments to call func with.
            is_enabled (bool, optional): True if the task is to be added to the scheduler.
        """
        self.name = name
        self.trigger = trigger
        self.id = id
        self.mistrigger_grace_time = mistrigger_grace_time
        self.coalesce = coalesce
        self.max_instances = max_instances
        self.next_run_time = next_run_time
        self.store = store
        self.executor = executor
        self.replace_existing = replace_existing
        self.args = args
        self.kwargs = kwargs
        self.is_enabled = is_enabled
        self.fn = fn

    def add_task(self, scheduler: SchedulerType) -> None:
        try:
            scheduler.add_task(
                self.fn,
                trigger=self.trigger,
                args=self.args,
                kwargs=self.kwargs,
                id=self.id,
                name=self.name,
                mistrigger_grace_time=self.mistrigger_grace_time,
                coalesce=self.coalesce,
                max_instances=self.max_instances,
                next_run_time=self.next_run_time,
                store=self.store,
                executor=self.executor,
                replace_existing=self.replace_existing,
            )
        except Exception as e:
            raise ImproperlyConfigured(str(e)) from e



from typing import TYPE_CHECKING, Any, Set, Type, TypeVar, cast

from pydantic import BaseModel, create_model
from pydantic.json_schema import models_json_schema

from esmerald.openapi.schemas import v3_1_0

if TYPE_CHECKING:
    from typing import Dict

REF_PREFIX = "#/components/schemas/"
SCHEMA_NAME_ATTRIBUTE = "__schema_name__"
MODE = "validation"

T = TypeVar("T", bound=v3_1_0.OpenAPI)


class OpenAPI310PydanticSchema(v3_1_0.Schema):
    """Special `Schema` class to indicate a reference from pydantic class."""

    schema_class: Type[BaseModel]
    """the class that is used for generate the schema"""


def construct_open_api_with_schema_class(
    open_api_schema: T,
) -> T:
    """Construct a new OpenAPI object, with the use of pydantic classes to
    produce JSON schemas.

    Args:
        open_api_schema: An instance of the OpenAPI model.

    Returns:
        new OpenAPI object with "#/components/schemas" values updated. If there is no update in
            "#/components/schemas" values, the original `open_api` will be returned.
    """
    copied_schema = open_api_schema.model_copy(deep=True)
    schema_classes = list(
        extract_pydantic_types_to_openapi_components(obj=copied_schema, ref_class=v3_1_0.Reference)
    )

    if not schema_classes:
        return open_api_schema

    if not copied_schema.components:
        copied_schema.components = v3_1_0.Components(schemas={})
    if copied_schema.components.schemas is None:  # pragma: no cover
        copied_schema.components.schemas = cast("Dict[str, Any]", {})

    schema_classes = [
        (
            (cls, MODE)  # type: ignore
            if not hasattr(cls, SCHEMA_NAME_ATTRIBUTE)
            else (create_model(getattr(cls, SCHEMA_NAME_ATTRIBUTE), __base__=cls), MODE)
        )
        for cls in schema_classes
    ]

    schema_classes.sort(key=lambda x: x[0].__name__)  # type: ignore
    _, json_schema = models_json_schema(schema_classes)  # type: ignore

    copied_schema.components.schemas.update(
        {
            key: v3_1_0.Schema.model_validate(schema_dict)
            for key, schema_dict in json_schema["$defs"].items()
        }
    )
    return copied_schema


def extract_pydantic_types_to_openapi_components(
    obj: Any, ref_class: Type[v3_1_0.Reference]
) -> Set[Type[BaseModel]]:
    """Recursively traverses the OpenAPI document, replacing any found Pydantic
    Models with $references to the schema's components section and returning
    the pydantic models themselves.

    Args:
        obj:
        ref_class:

    Returns:
        set of pydantic schema classes
    """
    pydantic_schemas: Set[Type[BaseModel]] = set()
    if isinstance(obj, BaseModel):
        fields = obj.model_fields_set
        for field in fields:
            child_obj = getattr(obj, field)
            if isinstance(child_obj, OpenAPI310PydanticSchema):
                setattr(obj, field, ref_class(ref=create_ref_prefix(child_obj.schema_class)))  # type: ignore
                pydantic_schemas.add(child_obj.schema_class)
            else:
                pydantic_schemas.update(
                    extract_pydantic_types_to_openapi_components(child_obj, ref_class=ref_class)
                )
    elif isinstance(obj, list):
        for index, elem in enumerate(obj):
            if isinstance(elem, OpenAPI310PydanticSchema):
                obj[index] = ref_class(ref=create_ref_prefix(elem.schema_class))  # type: ignore
                pydantic_schemas.add(elem.schema_class)
            else:
                pydantic_schemas.update(
                    extract_pydantic_types_to_openapi_components(elem, ref_class=ref_class)
                )
    elif isinstance(obj, dict):
        for key, value in obj.items():
            if isinstance(value, OpenAPI310PydanticSchema):
                obj[key] = ref_class(ref=create_ref_prefix(value.schema_class))  # type: ignore
                pydantic_schemas.add(value.schema_class)
            else:
                pydantic_schemas.update(
                    extract_pydantic_types_to_openapi_components(value, ref_class=ref_class)
                )
    return pydantic_schemas


def create_ref_prefix(model: Type[BaseModel]) -> str:
    """

    Args:
        model: Pydantic model instance.

    Returns:
        A prefixed name.
    """
    return REF_PREFIX + getattr(model, SCHEMA_NAME_ATTRIBUTE, model.__name__)



from typing import Any, Dict, List, Optional, TypeVar, Union

from pydantic import DirectoryPath, validate_call
from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class TemplateProtocol(Protocol):  # pragma: no cover
    def make_response(self, **context: Optional[Dict[str, Any]]) -> str: ...


TP = TypeVar("TP", bound=TemplateProtocol, covariant=True)


@runtime_checkable
class TemplateEngineProtocol(Protocol[TP]):  # pragma: no cover
    @validate_call
    def __init__(
        self, directory: Union[DirectoryPath, List[DirectoryPath]], **kwargs: Any
    ) -> None: ...

    def get_template(self, template_name: str) -> TP: ...



from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Mapping,
    Type,
    TypeVar,
    Union,
    get_args,
)

from lilya.middleware import DefineMiddleware
from lilya.types import ASGIApp
from typing_extensions import Literal

from esmerald.background import BackgroundTask, BackgroundTasks
from esmerald.exceptions import MissingDependency
from esmerald.routing.gateways import WebSocketGateway
from esmerald.routing.router import Include

try:
    from asyncz.schedulers import AsyncIOScheduler
except ImportError:
    AsyncIOScheduler = Any  # type: ignore

try:
    from esmerald.config.template import TemplateConfig as TemplateConfig
except MissingDependency:
    TemplateConfig = Any  # type: ignore

if TYPE_CHECKING:
    from esmerald.applications import Application, Esmerald
    from esmerald.conf.global_settings import EsmeraldAPISettings
    from esmerald.datastructures import Cookie, ResponseHeader, State as State
    from esmerald.injector import Inject
    from esmerald.protocols.middleware import MiddlewareProtocol as MiddlewareProtocol
    from esmerald.requests import Request
    from esmerald.responses import Response
    from esmerald.routing.apis.base import View
    from esmerald.routing.gateways import Gateway, WebhookGateway
    from esmerald.routing.router import (
        HTTPHandler as HTTPHandler,
        Router,
        WebSocketHandler as WebSocketHandler,
    )
    from esmerald.websockets import WebSocket
else:
    HTTPHandler = Any
    Application = Any
    Message = Any
    Receive = Any
    Scope = Any
    Send = Any
    BaseHTTPMiddleware = Any
    Inject = Any
    Request = Any
    WebSocket = Any
    State = Any
    Response = Any
    ResponseHeader = Any
    Cookie = Any
    Router = Any
    MiddlewareProtocol = Any
    View = Any
    Gateway = Any
    WebhookGateway = Any
    Esmerald = Any

AsyncAnyCallable = Callable[..., Awaitable[Any]]
HTTPMethod = Literal["GET", "POST", "DELETE", "PATCH", "PUT", "HEAD"]
HTTPMethods = get_args(HTTPMethod)

Middleware = DefineMiddleware

ResponseType = Type[Response]

Dependencies = Dict[str, Inject]

ExceptionType = TypeVar("ExceptionType", bound=Exception)
ExceptionHandler = Callable[[Request, ExceptionType], Response]
ExceptionHandlerMap = Mapping[Union[int, Type[Exception]], ExceptionHandler]

_ReservedKwargs = Literal[
    "request", "socket", "headers", "query", "cookies", "state", "data", "payload"
]

ReservedKwargs = get_args(_ReservedKwargs)

ResponseHeaders = Dict[str, ResponseHeader]
ResponseCookies = List[Cookie]
AsyncAnyCallable = Callable[..., Awaitable[Any]]  # type: ignore


SchedulerType = AsyncIOScheduler
DatetimeType = TypeVar("DatetimeType", bound=datetime)

ParentType = Union[View, Router, Gateway, WebSocketGateway, Esmerald, Include, Application]
APIGateHandler = Union[
    Gateway,
    WebSocketGateway,
]

RouteParent = Union[
    "Router", "Include", "ASGIApp", "Gateway", "WebSocketGateway", "WebhookGateway"
]

BackgroundTaskType = Union[BackgroundTask, BackgroundTasks]
SecurityScheme = Dict[str, List[str]]

ConnectionType = Union["Request", "WebSocket"]
DictStr = Dict[str, str]
DictAny = Dict[str, Any]
SettingsType = Type["EsmeraldAPISettings"]

LifeSpanHandler = Callable[[], Response]



from typing import Any, Callable, Iterable, Type, cast

import msgspec
from msgspec import ValidationError
from pydantic._internal._schema_generation_shared import (
    GetJsonSchemaHandler as GetJsonSchemaHandler,
)
from pydantic_core.core_schema import (
    CoreSchema,
    PlainValidatorFunctionSchema,
    with_info_plain_validator_function,
)

REF_TEMPLATE = "#/components/schemas/{name}"


class Struct(msgspec.Struct):
    """
    `Struct` object is the `esmerald.datastructures.msgspec` integration
    with Pydantic.

    When you want to integrate msgspec with your existing Pydantic models,
    you can simply use this datastructure directly to integrate in cleaner
    way. This will provide all the necessary data for Pydantic to generate
    the necessary schema data.

    This datastructure subclasses the `msgspec.Struct` directly which means
    there is no additional unique information added besides the
    Pydantic unique functionalities.

    **Example**

    ```python
    from pydantic import BaseModel

    from esmerald.datastructures.msgspec import Struct

    class User(Struct):
        name: str
        email: Union[str, None] = None


    class BaseUser(BaseModel):
        model_config = {"arbitrary_types_allowed": True}
        user: User

    ```
    """

    @classmethod
    def __get_validators__(cls: Type["Struct"]) -> Iterable[Callable[..., Any]]:
        yield cls.validate

    @classmethod
    def validate(cls: Type["Struct"], v: Any) -> Any:
        if not isinstance(v, msgspec.Struct):
            raise ValueError(f"Expected `esmerald.datastructures.msgspec.Struct`, got: {type(v)}")
        return v

    @classmethod
    def _validate(cls, __input_value: Any, _: Any) -> "Struct":
        try:
            return msgspec.json.decode(msgspec.json.encode(__input_value), type=cls)
        except ValidationError as e:
            raise e

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> Any:
        _, schema_definitions = msgspec.json.schema_components(
            types=(cls,), ref_template=REF_TEMPLATE
        )
        return schema_definitions[cls.__name__]

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Type[Any], handler: Callable[[Any], CoreSchema]
    ) -> CoreSchema:
        return cast(
            PlainValidatorFunctionSchema,
            with_info_plain_validator_function(cls._validate),
        )



from typing import Any, TypeVar

import bson
from lilya.types import ASGIApp
from mongoz import DocumentNotFound

from esmerald.config.jwt import JWTConfig
from esmerald.contrib.auth.common.middleware import CommonJWTAuthMiddleware
from esmerald.exceptions import AuthenticationError, NotAuthorized

T = TypeVar("T")

IDS = ["id", "pk", "_id"]


class JWTAuthMiddleware(CommonJWTAuthMiddleware):
    """
    The simple JWT authentication Middleware.
    """

    def __init__(
        self,
        app: "ASGIApp",
        config: "JWTConfig",
        user_model: T,
    ):
        super().__init__(app, config, user_model)
        """
        The user is simply the class type to be queried from the Saffier ORM.

        Example how to use:

            1. User table

                from esmerald.contrib.auth.saffier.base_user import User as BaseUser

                class User(BaseUser):
                    ...

            2. Middleware

                from esmerald.contrib.auth.saffier.middleware import JWTAuthMiddleware
                from esmerald.config import JWTConfig

                jwt_config = JWTConfig(...)

                class CustomJWTMidleware(JWTAuthMiddleware):
                    def __init__(self, app: "ASGIApp"):
                        super().__init__(app, config=jwt_config, user=User)

            3. The application
                from esmerald import Esmerald
                from myapp.middleware import CustomJWTMidleware

                app = Esmerald(routes=[...], middleware=[CustomJWTMidleware])

        """
        self.app = app
        self.config = config
        self.user_model: T = user_model  # type: ignore

    async def retrieve_user(self, token_sub: Any) -> T:
        """
        Retrieves a user from the database using the given token id.
        """
        try:
            sub = int(token_sub)
            token_sub = sub
        except (TypeError, ValueError):
            ...  # noqa

        if isinstance(token_sub, str) and self.config.user_id_field in IDS:
            token_sub = bson.ObjectId(token_sub)

        user_field = {self.config.user_id_field: token_sub}
        try:
            return await self.user_model.objects.get(**user_field)  # type: ignore
        except DocumentNotFound:
            raise NotAuthorized() from None
        except Exception as e:
            raise AuthenticationError(detail=str(e)) from e



from typing import TYPE_CHECKING, Any, Dict, List, Union

from esmerald.exceptions import MissingDependency, TemplateNotFound
from esmerald.protocols.template import TemplateEngineProtocol

if TYPE_CHECKING:  # pragma: no cover
    from pydantic import DirectoryPath

try:
    from mako.exceptions import TemplateLookupException as MakoTemplateNotFound
    from mako.lookup import TemplateLookup
    from mako.template import Template as MakoTemplate
except ImportError as exc:  # pragma: no cover
    raise MissingDependency("mako is not installed") from exc


class MakoTemplateEngine(TemplateEngineProtocol[MakoTemplate]):
    def __init__(
        self,
        directory: Union["DirectoryPath", List["DirectoryPath"]],
        env: Union[Any, None] = None,
        **env_options: Dict[Any, Any],
    ) -> None:
        super().__init__(directory)
        self.engine = TemplateLookup(
            directories=directory if isinstance(directory, (list, tuple)) else [directory]
        )
        self.options = env_options
        self.env = env

    def get_template(self, template_name: str) -> MakoTemplate:  # pragma: no cover
        try:
            return self.engine.get_template(template_name)
        except MakoTemplateNotFound as e:
            raise TemplateNotFound(name=template_name) from e



import inspect
from functools import cached_property
from typing import TYPE_CHECKING, Any, Dict, List, Union, _GenericAlias, cast, get_args

from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo

from esmerald.encoders import ENCODER_TYPES, is_body_encoder
from esmerald.enums import EncodingType
from esmerald.openapi.params import ResponseParam
from esmerald.params import Body
from esmerald.utils.constants import DATA, PAYLOAD
from esmerald.utils.models import create_field_model

if TYPE_CHECKING:
    from esmerald.routing.router import HTTPHandler, WebhookHandler


def get_base_annotations(base_annotation: Any) -> Dict[str, Any]:
    """
    Returns the annotations of the base class.

    Args:
        base (Any): The base class.

    Returns:
        Dict[str, Any]: The annotations of the base class.
    """
    base_annotations: Dict[str, Any] = {}
    for base in base_annotation.__bases__:
        base_annotations.update(**get_base_annotations(base))
        if hasattr(base, "__annotations__"):
            for name, annotation in base.__annotations__.items():
                base_annotations[name] = annotation
    return base_annotations


def convert_annotation_to_pydantic_model(field_annotation: Any) -> Any:
    """
    Converts any annotation of the body into a Pydantic
    base model.

    This is used for OpenAPI representation purposes only.

    Esmerald will try internally to convert the model into a Pydantic BaseModel,
    this will serve as representation of the model in the documentation but internally,
    it will use the native type to validate the data being sent and parsed in the
    payload/data field.

    Encoders are not supported in the OpenAPI representation, this is because the encoders
    are unique to Esmerald and are not part of the OpenAPI specification. This is why
    we convert the encoders into a Pydantic model for OpenAPI representation purposes only.
    """
    annotation_args = get_args(field_annotation)
    if isinstance(field_annotation, _GenericAlias):
        annotations = tuple(convert_annotation_to_pydantic_model(arg) for arg in annotation_args)
        field_annotation.__args__ = annotations
        return field_annotation

    if (
        not isinstance(field_annotation, BaseModel)
        and any(encoder.is_type(field_annotation) for encoder in ENCODER_TYPES)
        and inspect.isclass(field_annotation)
    ):
        field_definitions: Dict[str, Any] = {}

        # Get any possible annotations from the base classes
        # This can be useful for inheritance with custom encoders
        base_annotations: Dict[str, Any] = {**get_base_annotations(field_annotation)}
        field_annotations = {
            **base_annotations,
            **field_annotation.__annotations__,
        }
        for name, annotation in field_annotations.items():
            field_definitions[name] = (annotation, ...)
        return create_model(
            field_annotation.__name__,
            __config__={"arbitrary_types_allowed": True},
            **field_definitions,
        )
    return field_annotation


def get_original_data_field(
    handler: Union["HTTPHandler", "WebhookHandler", Any]
) -> Any:  # pragma: no cover
    """
    The field used for the payload body.

    This builds a model for the required data field. Validates the type of encoding
    being passed and builds a model if a datastructure is evaluated.
    """
    if (
        DATA in handler.signature_model.model_fields
        or PAYLOAD in handler.signature_model.model_fields
    ):
        data_or_payload = DATA if DATA in handler.signature_model.model_fields else PAYLOAD
        data = handler.signature_model.model_fields[data_or_payload]

        if not isinstance(data, Body):
            body = Body(alias="body")
            for key, _ in data._attributes_set.items():
                setattr(body, key, getattr(data, key, None))
        else:
            body = data

        # Check the annotation type
        body.annotation = convert_annotation_to_pydantic_model(body.annotation)

        if not body.title:
            body.title = f"Body_{handler.operation_id}"

        # For everything else that is not MULTI_PART
        extra = cast("Dict[str, Any]", body.json_schema_extra) or {}
        if extra.get("media_type", EncodingType.JSON) != EncodingType.MULTI_PART:
            return body

        # For Uploads and Multi Part
        args = get_args(body.annotation)
        name = "File" if not args else "Files"

        model = create_field_model(field=body, name=name, model_name=body.title)
        data_field = Body(annotation=model, title=body.title)

        for key, _ in data._attributes_set.items():
            if key != "annotation":
                setattr(data_field, key, getattr(body, key, None))

        return data_field


def get_complex_data_field(
    handler: Union["HTTPHandler", "WebhookHandler", Any], fields: Dict[str, FieldInfo]
) -> Any:  # pragma: no cover
    """
    The field used for the payload body.

    This builds a model for the required data field. Validates the type of encoding
    being passed and builds a model if a datastructure is evaluated.
    """
    body_fields_set = set()
    body_fields: Dict[str, FieldInfo] = {}

    for name, field in fields.items():
        if name in body_fields_set:
            continue

        body_fields_set.add(name)
        body_fields[name] = field

    # Set the field definitions
    field_definitions = {}
    for name, param in body_fields.items():
        param.annotation = convert_annotation_to_pydantic_model(param.annotation)
        field_definitions[name] = param.annotation, ...

    # Create the model from the field definitions
    model = create_model(  # type: ignore
        "DataField", __config__={"arbitrary_types_allowed": True}, **field_definitions
    )
    # Create the body field
    body = Body(annotation=model, title=f"Body_{handler.operation_id}")

    # Check the annotation type
    if not body.title:
        body.title = f"Body_{handler.operation_id}"
    return body


def get_data_field(handler: Union["HTTPHandler", "WebhookHandler", Any]) -> Any:
    """
    Retrieves the data field from the given handler.

    Args:
        handler (Union[HTTPHandler, WebhookHandler, Any]): The handler object.

    Returns:
        Any: The data field.

    Raises:
        None

    This function checks if the handler has any body encoder fields. If there are less than 2 body encoder fields,
    it calls the get_original_data_field function. Otherwise, it calls the get_complex_data_field function.

    One the steps is to make sure backwards compatibility and this means to make sure previous
    versions of Esmerald are supported the way they are supposed to.

    The other thing is to make sure we extract any value from the signature of the handler and
    match against any encoder, custom or default, and isolate them as body fields and then extract
    the values that are not in the dependencies since those are not considered part of the body
    but as dependency itself.

    If the body fields are less than 1 and using the reserved `data` or `payload` then it will
    default to the normal Esmerald processing, otherwise it will use the complex approach of
    designing the OpenAPI body.
    """
    is_data_or_payload = (
        DATA
        if DATA in handler.signature_model.model_fields
        else (PAYLOAD if PAYLOAD in handler.signature_model.model_fields else None)
    )

    # If there are no body fields, we simply return the original
    # default Esmerald body parsing
    if not handler.body_encoder_fields:
        return get_original_data_field(handler)

    if len(handler.body_encoder_fields) < 2 and is_data_or_payload is not None:
        return get_original_data_field(handler)
    return get_complex_data_field(handler, fields=handler.body_encoder_fields)


class OpenAPIFieldInfoMixin:
    """
    Used for validating model fields necessary for the
    OpenAPI parsing only.

    Don't use this anywhere else.
    """

    @property
    def body_encoder_fields(self) -> Dict[str, FieldInfo]:
        """
        The fields that are body encoders.

        This is used for OpenAPI representation purposes only.
        """
        # Making sure the dependencies are not met as body fields for OpenAPI representation
        handler_dependencies = set(self.get_dependencies().keys())

        # Getting everything else that is not considered a dependency
        body_encoder_fields = {
            name: field
            for name, field in self.signature_model.model_fields.items()
            if is_body_encoder(field.annotation) and name not in handler_dependencies
        }
        return body_encoder_fields

    @cached_property
    def response_models(self) -> Dict[int, Any]:
        """
        The models converted into pydantic fields with the model used for OpenAPI.

        The response models can be a list representation or a single object representation.
        If another type of object is passed through the `model`, an Assertation error is raised.
        """
        responses: Dict[int, ResponseParam] = {}
        if self.responses:
            for status_code, response in self.responses.items():
                model = response.model[0] if isinstance(response.model, list) else response.model

                annotation = (
                    List[model] if isinstance(response.model, list) else model  # type: ignore
                )

                responses[status_code] = ResponseParam(
                    annotation=convert_annotation_to_pydantic_model(annotation),
                    description=response.description,
                    alias=model.__name__,
                )
        return responses

    @cached_property
    def data_field(self) -> Any:  # pragma: no cover
        """
        The field used for the payload body.

        This builds a model for the required data field. Validates the type of encoding
        being passed and builds a model if a datastructure is evaluated.
        """
        data_field = get_data_field(self)
        return data_field



from abc import ABC, abstractmethod
from typing import Any, Set

from lilya._internal._connection import Connection
from lilya.types import ASGIApp, Receive, Scope, Send
from typing_extensions import Annotated, Doc

from esmerald.enums import ScopeType
from esmerald.parsers import ArbitraryBaseModel
from esmerald.protocols.middleware import MiddlewareProtocol


class AuthResult(ArbitraryBaseModel):
    user: Annotated[
        Any,
        Doc(
            """
            Arbitrary user coming from the `authenticate` of the `BaseAuthMiddleware`
            and can be assigned to the `request.user`.
            """
        ),
    ]


class BaseAuthMiddleware(ABC, MiddlewareProtocol):  # pragma: no cover
    """
    `BaseAuthMiddleware` is the object that you can implement if you
    want to implement any `authentication` middleware with Esmerald.

    It is not mandatory to use it and you are free to implement your.

    Esmerald being based on Lilya, also offers a simple but powerful
    interface for handling `authentication` and [permissions](https://esmerald.dev/permissions/).

    Once you have installed the `AuthenticationMiddleware` and implemented the
    `authenticate`, the `request.user` will be available in any of your
    endpoints.

    Read more about how [Esmerald implements](https://esmerald.dev/middleware/middleware#baseauthmiddleware) the `BaseAuthMiddleware`.

    When implementing the `authenticate`, you must assign the result into the
    `AuthResult` object in order for the middleware to assign the `request.user`
    properly.

    The `AuthResult` is of type `esmerald.middleware.authentication.AuthResult`.
    """

    def __init__(
        self,
        app: Annotated[
            ASGIApp,
            Doc(
                """
                An ASGI type callable.
                """
            ),
        ],
    ):
        super().__init__(app)
        self.app = app
        self.scopes: Set[str] = {ScopeType.HTTP, ScopeType.WEBSOCKET}

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Function callable that automatically will call the `authenticate` function
        from any middleware subclassing the `BaseAuthMiddleware` and assign the `AuthUser` user
        to the `request.user`.
        """
        if scope["type"] not in self.scopes:
            await self.app(scope, receive, send)
            return

        auth_result = await self.authenticate(Connection(scope))
        scope["user"] = auth_result.user
        await self.app(scope, receive, send)

    @abstractmethod
    async def authenticate(self, request: Connection) -> AuthResult:
        """
        The abstract method that needs to be implemented for any authentication middleware.
        """
        raise NotImplementedError("authenticate must be implemented.")



from datetime import datetime
from typing import Any, Dict, Optional, Union, cast

from asyncz.triggers.types import TriggerType
from asyncz.typing import Undefined, undefined

from esmerald.contrib.schedulers.asyncz.config import Task


def scheduler(
    *,
    name: Optional[str] = None,
    trigger: Optional[TriggerType] = None,
    id: Optional[str] = None,
    mistrigger_grace_time: Union[int, Undefined, None] = undefined,
    coalesce: Union[bool, Undefined] = undefined,
    max_instances: Union[int, Undefined, None] = undefined,
    next_run_time: Union[datetime, str, Undefined, None] = undefined,
    store: Union[str, None] = None,
    executor: Union[str, None] = None,
    replace_existing: bool = True,
    extra_args: Optional[Any] = None,
    extra_kwargs: Optional[Dict[str, Any]] = None,
    is_enabled: bool = True,
) -> "Task":
    def wrapper(func: Any) -> Task:
        task = Task(
            fn=func,
            name=name,
            trigger=trigger,
            id=id,
            mistrigger_grace_time=mistrigger_grace_time,
            coalesce=coalesce,
            max_instances=max_instances,
            next_run_time=next_run_time,
            store=store,
            executor=executor,
            replace_existing=replace_existing,
            args=extra_args,
            kwargs=extra_kwargs,
            is_enabled=is_enabled,
        )
        return task

    return cast(Task, wrapper)



from typing import Any

from pydantic.fields import FieldInfo

from esmerald.utils.constants import IS_DEPENDENCY, SKIP_VALIDATION


def is_dependency_field(val: Any) -> bool:
    json_schema_extra = getattr(val, "json_schema_extra", None) or {}
    return bool(isinstance(val, FieldInfo) and bool(json_schema_extra.get(IS_DEPENDENCY)))


def should_skip_dependency_validation(val: Any) -> bool:
    json_schema_extra = getattr(val, "json_schema_extra", None) or {}
    return bool(is_dependency_field(val) and json_schema_extra.get(SKIP_VALIDATION))



from typing import Optional, Union

from pydantic import AnyUrl, BaseModel, ConfigDict, EmailStr


class Contact(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        json_schema_extra={
            "examples": [
                {
                    "name": "API Support",
                    "url": "http://www.example.com/support",
                    "email": "support@example.com",
                }
            ]
        },
    )
    """Contact information for the exposed API."""

    name: Optional[str] = None
    """
    The identifying name of the contact person/organization.
    """

    url: Optional[AnyUrl] = None
    """
    The URL pointing to the contact information.
    MUST be in the form of a URL.
    """

    email: Optional[Union[EmailStr, str]] = None
    """
    The email address of the contact person/organization.
    MUST be in the form of an email address.
    """



from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Type, Union

from esmerald.core.di.provider import load_provider
from esmerald.parsers import ArbitraryHashableBaseModel
from esmerald.transformers.signature import SignatureModel
from esmerald.typing import Void
from esmerald.utils.helpers import is_async_callable

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.typing import AnyCallable


class Factory:
    def __init__(self, provides: Union["AnyCallable", str], *args: Any) -> None:
        """
        The provider can be passed in separate ways. Via direct callable
        or via string value where it will be automatically imported by the application.
        """
        self.__args: Tuple[Any, ...] = ()
        self.set_args(*args)
        self.is_nested: bool = False

        if isinstance(provides, str):
            self.provides, self.is_nested = load_provider(provides)
        else:
            self.provides = provides

    def set_args(self, *args: Any) -> None:
        self.__args = args

    @property
    def cls(self) -> "AnyCallable":  # pragma: no cover
        """Return provided type."""
        return self.provides

    async def __call__(self) -> Any:
        """
        This handles with normal and nested imports.

        Example:

            1. MyClass.func
            2. MyClass.AnotherClass.func
        """
        if self.is_nested:
            self.provides = self.provides()

        if is_async_callable(self.provides):
            value = await self.provides(*self.__args)
        else:
            value = self.provides(*self.__args)
        return value


class Inject(ArbitraryHashableBaseModel):
    def __init__(self, dependency: "AnyCallable", use_cache: bool = False, **kwargs: Any):
        super().__init__(**kwargs)
        self.dependency = dependency
        self.signature_model: Optional["Type[SignatureModel]"] = None
        self.use_cache = use_cache
        self.value: Any = Void

    async def __call__(self, **kwargs: Dict[str, Any]) -> Any:
        if self.use_cache and self.value is not Void:
            return self.value

        if is_async_callable(self.dependency):
            value = await self.dependency(**kwargs)
        else:
            value = self.dependency(**kwargs)

        if self.use_cache:
            self.value = value

        return value

    def __eq__(self, other: Any) -> bool:
        return other is self or (
            isinstance(other, self.__class__)
            and other.dependency == self.dependency
            and other.use_cache == self.use_cache
            and other.value == self.value
        )



import os
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Optional, Type

from lilya._internal._module_loading import import_string

from esmerald.utils.functional import LazyObject, empty

if TYPE_CHECKING:
    from esmerald.conf.global_settings import EsmeraldAPISettings

ENVIRONMENT_VARIABLE = "ESMERALD_SETTINGS_MODULE"


@lru_cache
def reload_settings() -> Type["EsmeraldAPISettings"]:
    """
    Reloads the global settings.
    """
    settings_module: str = os.environ.get(
        ENVIRONMENT_VARIABLE, "esmerald.conf.global_settings.EsmeraldAPISettings"
    )
    settings: Type["EsmeraldAPISettings"] = import_string(settings_module)
    return settings


class EsmeraldLazySettings(LazyObject):
    """
    A lazy proxy for either global Esmerald settings or a custom settings object.
    The user can manually configure settings prior to using them. Otherwise,
    Esmerald uses the settings module pointed to by ESMERALD_SETTINGS_MODULE.
    """

    def _setup(self, name: Optional[str] = None) -> None:
        """
        Load the settings module pointed to by the environment variable. This
        is used the first time settings are needed, if the user hasn't
        configured settings manually.
        """
        settings: Type["EsmeraldAPISettings"] = reload_settings()

        for setting, _ in settings().model_dump().items():
            assert setting.islower(), "%s should be in lowercase." % setting

        self._wrapped = settings()

    def configure(self, override_settings: Type["EsmeraldAPISettings"]) -> None:
        """
        Making sure the settings are overriden by the settings_module
        provided by a given application and therefore use it as the application
        global.
        """
        self._wrapped = override_settings

    def __repr__(self: "EsmeraldLazySettings") -> str:
        # Hardcode the class name as otherwise it yields 'Settings'.
        if self._wrapped is empty:
            return "<EsmeraldLazySettings [Unevaluated]>"
        return '<EsmeraldLazySettings "{settings_module}">'.format(
            settings_module=self._wrapped.__class__.__name__
        )

    @property
    def configured(self) -> Any:
        """Return True if the settings have already been configured."""
        return self._wrapped is not empty


settings: Type["EsmeraldAPISettings"] = EsmeraldLazySettings()  # type: ignore
__lazy_settings__: Type["EsmeraldAPISettings"] = EsmeraldLazySettings()  # type: ignore



import sys
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Union,
    cast,
)

from httpx._client import CookieTypes
from lilya.conf.context_vars import set_override_settings
from lilya.testclient import TestClient  # noqa
from pydantic import AnyUrl

from esmerald.applications import Esmerald
from esmerald.conf import settings
from esmerald.conf.global_settings import EsmeraldAPISettings as Settings
from esmerald.contrib.schedulers import SchedulerConfig
from esmerald.encoders import Encoder
from esmerald.openapi.schemas.v3_1_0 import Contact, License, SecurityScheme
from esmerald.utils.crypto import get_random_secret_key

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

P = ParamSpec("P")


if TYPE_CHECKING:  # pragma: no cover
    from typing_extensions import Literal

    from esmerald.config import (
        CORSConfig,
        CSRFConfig,
        OpenAPIConfig,
        SessionConfig,
        StaticFilesConfig,
    )
    from esmerald.interceptors.types import Interceptor
    from esmerald.permissions.types import Permission
    from esmerald.pluggables import Pluggable
    from esmerald.routing.gateways import WebhookGateway
    from esmerald.types import (
        APIGateHandler,
        Dependencies,
        ExceptionHandlerMap,
        LifeSpanHandler,
        Middleware,
        SettingsType,
        TemplateConfig,
    )


class EsmeraldTestClient(TestClient):  # type: ignore
    app: Esmerald

    def __init__(
        self,
        app: Esmerald,
        base_url: str = "http://testserver",
        raise_server_exceptions: bool = True,
        root_path: str = "",
        backend: "Literal['asyncio', 'trio']" = "asyncio",
        backend_options: Optional[Dict[str, Any]] = None,
        cookies: Optional[CookieTypes] = None,
        headers: Dict[str, str] = None,
    ):
        super().__init__(
            app=app,
            base_url=base_url,
            raise_server_exceptions=raise_server_exceptions,
            root_path=root_path,
            backend=backend,
            backend_options=backend_options,
            cookies=cookies,
            headers=headers,
        )

    def __enter__(self, *args: Any, **kwargs: Dict[str, Any]) -> "EsmeraldTestClient":
        return cast("EsmeraldTestClient", super().__enter__(*args, **kwargs))


def create_client(
    routes: Union["APIGateHandler", List["APIGateHandler"]],
    *,
    settings_module: Union[Optional["SettingsType"], Optional[str]] = None,
    debug: Optional[bool] = None,
    app_name: Optional[str] = None,
    title: Optional[str] = None,
    version: Optional[str] = None,
    summary: Optional[str] = None,
    description: Optional[str] = None,
    contact: Optional[Contact] = None,
    terms_of_service: Optional[AnyUrl] = None,
    license: Optional[License] = None,
    security: Optional[List[SecurityScheme]] = None,
    servers: Optional[List[Dict[str, Union[str, Any]]]] = None,
    secret_key: Optional[str] = get_random_secret_key(),
    allowed_hosts: Optional[List[str]] = None,
    allow_origins: Optional[List[str]] = None,
    base_url: str = "http://testserver",
    backend: "Literal['asyncio', 'trio']" = "asyncio",
    backend_options: Optional[Dict[str, Any]] = None,
    interceptors: Optional[List["Interceptor"]] = None,
    pluggables: Optional[Dict[str, "Pluggable"]] = None,
    permissions: Optional[List["Permission"]] = None,
    dependencies: Optional["Dependencies"] = None,
    middleware: Optional[List["Middleware"]] = None,
    csrf_config: Optional["CSRFConfig"] = None,
    exception_handlers: Optional["ExceptionHandlerMap"] = None,
    openapi_config: Optional["OpenAPIConfig"] = None,
    on_shutdown: Optional[List["LifeSpanHandler"]] = None,
    on_startup: Optional[List["LifeSpanHandler"]] = None,
    cors_config: Optional["CORSConfig"] = None,
    session_config: Optional["SessionConfig"] = None,
    scheduler_config: Optional[SchedulerConfig] = None,
    enable_scheduler: bool = None,
    enable_openapi: bool = True,
    include_in_schema: bool = True,
    openapi_version: Optional[str] = "3.1.0",
    raise_server_exceptions: bool = True,
    root_path: str = "",
    static_files_config: Optional["StaticFilesConfig"] = None,
    template_config: Optional["TemplateConfig"] = None,
    lifespan: Optional[Callable[["Esmerald"], "AsyncContextManager"]] = None,
    cookies: Optional[CookieTypes] = None,
    redirect_slashes: Optional[bool] = None,
    tags: Optional[List[str]] = None,
    webhooks: Optional[Sequence["WebhookGateway"]] = None,
    encoders: Optional[Sequence[Encoder]] = None,
) -> EsmeraldTestClient:
    return EsmeraldTestClient(
        app=Esmerald(
            settings_module=settings_module,
            debug=debug,
            title=title,
            version=version,
            summary=summary,
            description=description,
            contact=contact,
            terms_of_service=terms_of_service,
            license=license,
            security=security,
            servers=servers,
            routes=cast("Any", routes if isinstance(routes, list) else [routes]),
            app_name=app_name,
            secret_key=secret_key,
            allowed_hosts=allowed_hosts,
            allow_origins=allow_origins,
            interceptors=interceptors,
            permissions=permissions,
            dependencies=dependencies,
            middleware=middleware,
            csrf_config=csrf_config,
            exception_handlers=exception_handlers,
            openapi_config=openapi_config,
            on_shutdown=on_shutdown,
            on_startup=on_startup,
            cors_config=cors_config,
            scheduler_config=scheduler_config,
            enable_scheduler=enable_scheduler,
            static_files_config=static_files_config,
            template_config=template_config,
            session_config=session_config,
            lifespan=lifespan,
            redirect_slashes=redirect_slashes,
            enable_openapi=enable_openapi,
            openapi_version=openapi_version,
            include_in_schema=include_in_schema,
            tags=tags,
            webhooks=webhooks,
            pluggables=pluggables,
            encoders=encoders,
        ),
        base_url=base_url,
        backend=backend,
        backend_options=backend_options,
        root_path=root_path,
        raise_server_exceptions=raise_server_exceptions,
        cookies=cookies,
    )


class override_settings:
    """
    A context manager that allows overriding Esmerald settings temporarily.

    Usage:
    ```
    with override_settings(SETTING_NAME=value):
        # code that uses the overridden settings
    ```

    The `override_settings` class can also be used as a decorator.

    Usage:
    ```
    @override_settings(SETTING_NAME=value)
    def test_function():
        # code that uses the overridden settings
    ```
    """

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize the class with the given keyword arguments.

        Args:
            **kwargs: Additional keyword arguments to be stored as options.

        Returns:
            None
        """
        self.app = kwargs.pop("app", None)
        self.options = kwargs

    def __enter__(self) -> None:
        """
        Enter the context manager and set the modified settings.

        Saves the original settings and sets the modified settings
        based on the provided options.

        Returns:
            None
        """
        self._original_settings = settings._wrapped
        settings._wrapped = Settings(settings._wrapped, **self.options)  # type: ignore
        set_override_settings(True)

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """
        Restores the original settings and sets them up again.

        Args:
            exc_type (Any): The type of the exception raised, if any.
            exc_value (Any): The exception instance raised, if any.
            traceback (Any): The traceback for the exception raised, if any.
        """
        settings._wrapped = self._original_settings
        settings._setup()
        set_override_settings(False)

    def __call__(self, test_func: Any) -> Any:
        """
        Decorator that wraps a test function and executes it within a context manager.

        Args:
            test_func (Any): The test function to be wrapped.

        Returns:
            Any: The result of the test function.

        """

        @wraps(test_func)
        def inner(*args: P.args, **kwargs: P.kwargs) -> Any:
            with self:
                return test_func(*args, **kwargs)

        return inner



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from typing_extensions import Annotated, Doc

from esmerald.exceptions import ImproperlyConfigured

if TYPE_CHECKING:
    from esmerald.requests import Request
    from esmerald.types import APIGateHandler

SAFE_METHODS = ("GET", "HEAD", "OPTIONS")


class BaseOperationHolder:  # pragma: no cover
    def __and__(self, other: Any) -> "OperandHolder":
        return OperandHolder(AND, self, other)

    def __or__(self, other: "Any") -> "OperandHolder":
        return OperandHolder(OR, self, other)

    def __rand__(self, other: Any) -> "OperandHolder":
        return OperandHolder(AND, other, self)

    def __ror__(self, other: "BasePermission") -> "OperandHolder":
        return OperandHolder(OR, other, self)

    def __invert__(self) -> "SingleOperand":
        return SingleOperand(NOT, self)


class SingleOperand(BaseOperationHolder):
    def __init__(self, operator_class: Any, op1_class: Any) -> None:
        self.operator_class = operator_class
        self.op1_class = op1_class

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        op1 = self.op1_class(*args, **kwargs)
        return self.operator_class(op1)


class OperandHolder(BaseOperationHolder):
    def __init__(self, operator_class: Any, op1_class: Any, op2_class: Any) -> None:
        self.operator_class = operator_class
        self.op1_class = op1_class
        self.op2_class = op2_class

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        op1 = self.op1_class(*args, **kwargs)
        op2 = self.op2_class(*args, **kwargs)
        return self.operator_class(op1, op2)


class AND:
    def __init__(self, op1: "BasePermission", op2: "BasePermission") -> None:
        self.op1 = op1
        self.op2 = op2

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        return self.op1.has_permission(request, apiview) and self.op2.has_permission(
            request, apiview
        )


class OR:
    def __init__(self, op1: "BasePermission", op2: "BasePermission") -> None:
        self.op1 = op1
        self.op2 = op2

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        return self.op1.has_permission(request, apiview) or self.op2.has_permission(
            request, apiview
        )


class NOT:
    def __init__(self, op1: "BasePermission") -> None:
        self.op1 = op1

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        return not self.op1.has_permission(request, apiview)


class BasePermissionMetaclass(BaseOperationHolder, type):  # type: ignore[misc,unused-ignore]
    ...


class BasePermission(metaclass=BasePermissionMetaclass):
    """
    `BasePermission` class object. The entry-point for **all** permissions
    used by `Esmerald.

    When creating a permission or a set of permissions for the application,
    those **must** inherit from the `BasePermission` and implement the `has_permission`.

    **Example**

    ```python
    from esmerald import BasePermission, Request
    from esmerald.types import APIGateHandler


    class IsProjectAllowed(BasePermission):
        '''
        Permission to validate if has access to a given project
        '''

        async def has_permission(self, request: "Request", apiview: "APIGateHandler"):
            allow_project = request.headers.get("allow_access")
            return bool(allow_project)
    ```
    """

    def has_permission(
        self,
        request: Annotated[
            "Request",
            Doc(
                """
                The request object being passed through the request.
                """
            ),
        ],
        apiview: Annotated[
            "APIGateHandler",
            Doc(
                """
                A [handler](https://esmerald.dev/routing/handlers/) usually
                corresponding the [level](https://esmerald.dev/application/levels/)
                where the permission is placed.
                """
            ),
        ],
    ) -> bool:
        """
        **Mandatory** functionality and entry-point for verifying
        if the resource is available or not.

        The `has_permission` can be both `sync` and `async` depending
        of the needs of application.

        Returns:
            Boolean indicating if has or not permission to access the specific resource.

        **Example with `async`**

        ```python
        from esmerald import BasePermission, Request
        from esmerald.types import APIGateHandler


        class IsProjectAllowed(BasePermission):
            '''
            Permission to validate if has access to a given project
            '''

            async def has_permission(self, request: "Request", apiview: "APIGateHandler") -> bool:
                allow_project = request.headers.get("allow_access")
                return bool(allow_project)
        ```

        **Example with `sync`**

        ```python
        from esmerald import BasePermission, Request
        from esmerald.types import APIGateHandler


        class IsProjectAllowed(BasePermission):
            '''
            Permission to validate if has access to a given project
            '''

            def has_permission(self, request: "Request", apiview: "APIGateHandler") -> bool:
                allow_project = request.headers.get("allow_access")
                return bool(allow_project)
        ```
        """
        return True


class BaseAbstractUserPermission(ABC):
    """
    Abstract Base class for user validation permissions.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        try:
            return hasattr(request, "user")
        except ImproperlyConfigured:
            return False

    @abstractmethod
    def is_user_authenticated(self, request: "Request") -> bool:
        """
        This method must be overridden by subclasses.

        Args:
            request: A Lilya 'Connection' instance.

        Returns:
            bool: True or False
        """
        raise NotImplementedError("is_user_uthenticated() must be implemented.")

    @abstractmethod
    def is_user_staff(self, request: "Request") -> bool:
        """
        This method must be overridden by subclasses.

        Args:
            request: A Lilya 'Connection' instance.

        Returns:
            bool: True or False
        """
        raise NotImplementedError("is_user_staff() must be implemented.")


class AllowAny(BasePermission):
    """
    Allow any access.
    This isn't strictly required, since you could use an empty
    permission_classes list, but it's useful because it makes the intention
    more explicit.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        return True


class DenyAll(BasePermission):
    """
    Deny all access.
    This isn't strictly required, since you could use an empty
    permission_classes list, but it's useful because it makes the intention
    more explicit.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        return False


class IsAuthenticated(BaseAbstractUserPermission):
    """
    Allows access only to authenticated users.
    Raises exception if the AuthenticationMiddleware is not in the `middleware` settings.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        """
        Args:
            request: A Lilya 'Connection' instance.
            apiview: A Esmerald 'APIController' instance or a `APIGateHandler` instance.

        Returns:
            bool: True or False
        """
        super().has_permission(request, apiview)
        return bool(request.user and self.is_user_authenticated(request))


class IsAdminUser(BaseAbstractUserPermission):
    """
    Allows access only to admin users.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        """
        Args:
            request: A Lilya 'Connection' instance.
            apiview: A Esmerald 'APIController' instance or a `APIGateHandler` instance.

        Returns:
            bool: True or False
        """
        super().has_permission(request, apiview)
        return bool(request.user and self.is_user_staff(request))


class IsAuthenticatedOrReadOnly(BaseAbstractUserPermission):
    """
    The request is authenticated as a user, or is a read-only request.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        """
        Args:
            request: A Lilya 'Connection' instance.
            apiview: A Esmerald 'APIController' instance or a `APIGateHandler` instance.

        Returns:
            bool: True or False
        """
        super().has_permission(request, apiview)
        return bool(
            request.method in SAFE_METHODS or request.user and self.is_user_authenticated(request)
        )



from typing import Any, Dict, List, Optional, Sequence, Union

from pydantic import AnyUrl, BaseModel
from typing_extensions import Annotated, Doc

from esmerald.openapi.docs import (
    get_rapidoc_ui_html,
    get_redoc_html,
    get_stoplight_html,
    get_swagger_ui_html,
    get_swagger_ui_oauth2_redirect_html,
)
from esmerald.openapi.models import Contact, License
from esmerald.openapi.openapi import get_openapi
from esmerald.openapi.schemas.v3_1_0.security_scheme import SecurityScheme
from esmerald.requests import Request
from esmerald.responses import HTMLResponse, JSONResponse
from esmerald.routing.handlers import get


class OpenAPIConfig(BaseModel):
    """
    An instance of [OpenAPIConfig](https://esmerald.dev/configurations/openapi/config/).

    This object is then used by Esmerald to create the [OpenAPI](https://esmerald.dev/openapi/) documentation.

    **Note**: Here is where the defaults for Esmerald OpenAPI are overriden and if
    this object is passed, then the previous defaults of the settings are ignored.

    !!! Tip
        This is the way you could override the defaults all in one go
        instead of doing attribute by attribute.

    **Example**

    ```python
    from esmerald import OpenAPIConfig

    openapi_config = OpenAPIConfig(
        title="Black Window",
        openapi_url="/openapi.json",
        docs_url="/docs/swagger",
        redoc_url="/docs/redoc",
    )

    app = Esmerald(openapi_config=openapi_config)
    ```

    !!! Important
        Esmerald when starting an application, checks the attributes and looks for an
        `openapi_config` parameter.

        If the parameter is not specified, `Esmerald` will automatically use the internal
        settings system to generate the default OpenAPIConfig and populate the values.
    """

    title: Annotated[
        Optional[str],
        Doc(
            """
            Title of the application/API documentation.
            """
        ),
    ] = None
    version: Annotated[
        Optional[str],
        Doc(
            """
            The version of the API documentation.
            """
        ),
    ] = None
    summary: Annotated[
        Optional[str],
        Doc(
            """
            Simple and short summary text of the application/API.
            """
        ),
    ] = None
    description: Annotated[
        Optional[str],
        Doc(
            """
            A longer and more descriptive explanation of the application/API documentation.
            """
        ),
    ] = None
    contact: Annotated[
        Optional[Contact],
        Doc(
            """
            API contact information. This is an OpenAPI schema contact, meaning, in a dictionary format compatible with OpenAPI or an instance of `esmerald.openapi.schemas.v3_1_0.contact.Contact`.
            """
        ),
    ] = None
    terms_of_service: Annotated[
        Optional[AnyUrl],
        Doc(
            """
            URL to a page that contains terms of service.
            """
        ),
    ] = None
    license: Annotated[
        Optional[License],
        Doc(
            """
            API Licensing information. This is an OpenAPI schema licence, meaning, in a dictionary format compatible with OpenAPI or an instance of `esmerald.openapi.schemas.v3_1_0.license.License`.
            """
        ),
    ] = None
    security: Annotated[
        Optional[List[SecurityScheme]],
        Doc(
            """
            API Security requirements information. This is an OpenAPI schema security, meaning, in a dictionary format compatible with OpenAPI or an instance of `esmerald.openapi.schemas.v3_1_0.security_requirement.SecurityScheme`.
            """
        ),
    ] = None
    servers: Annotated[
        Optional[List[Dict[str, Union[str, Any]]]],
        Doc(
            """
            A python list with dictionary compatible with OpenAPI specification.
            """
        ),
    ] = None
    tags: Annotated[
        Optional[List[str]],
        Doc(
            """
            A list of OpenAPI compatible tag (string) information.
            """
        ),
    ] = None
    openapi_version: Annotated[
        Optional[str],
        Doc(
            """
            The version of the OpenAPI being used. Esmerald uses the version 3.1.0 by
            default and tis can be useful if you want to trick some of the existing tools
            that require a lower version.
            """
        ),
    ] = None
    openapi_url: Annotated[
        Optional[str],
        Doc(
            """
            URL of the `openapi.json` in the format of a path.

            Example: `/openapi.json.`
            """
        ),
    ] = None
    root_path_in_servers: Annotated[
        bool,
        Doc(
            """
            A `boolean` flag indicating if the root path should be included in the servers.
            """
        ),
    ] = True
    docs_url: Annotated[
        Optional[str],
        Doc(
            """
            String default relative URL where the Swagger documentation
            shall be accessed in the application.

            Example: '/docs/swagger`.
            """
        ),
    ] = None
    redoc_url: Annotated[
        Optional[str],
        Doc(
            """
            String default relative URL where the ReDoc documentation
            shall be accessed in the application.

            Example: '/docs/swagger`.
            """
        ),
    ] = None
    swagger_ui_oauth2_redirect_url: Annotated[
        Optional[str],
        Doc(
            """
            String default relative URL where the Swagger UI OAuth Redirect URL
            shall be accessed in the application.

            Example: `/docs/oauth2-redirect`.
            """
        ),
    ] = None
    redoc_js_url: Annotated[
        Optional[str],
        Doc(
            """
            String default URL where the ReDoc Javascript is located
            and used within OpenAPI documentation,

            Example: `https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js`.
            """
        ),
    ] = None
    redoc_favicon_url: Annotated[
        Optional[str],
        Doc(
            """
            String default URL where the ReDoc favicon is located
            and used within OpenAPI documentation,

            Example: `https://esmerald.dev/statics/images/favicon.ico`.
            """
        ),
    ] = None
    swagger_ui_init_oauth: Annotated[
        Optional[Dict[str, Any]],
        Doc(
            """
            String default relative URL where the Swagger Init Auth documentation
            shall be accessed in the application.
            """
        ),
    ] = None
    swagger_ui_parameters: Annotated[
        Optional[Dict[str, Any]],
        Doc(
            """
            A mapping with parameters to be passed onto Swagger.
            """
        ),
    ] = None
    swagger_js_url: Annotated[
        Optional[str],
        Doc(
            """
            Boolean flag indicating if the google fonts shall be used
            in the ReDoc OpenAPI Documentation.
            """
        ),
    ] = None
    swagger_css_url: Annotated[
        Optional[str],
        Doc(
            """
            String default URL where the Swagger Javascript is located
            and used within OpenAPI documentation,

            Example: `https://cdn.jsdelivr.net/npm/swagger-ui-dist@5.1.3/swagger-ui-bundle.min.js`.
            """
        ),
    ] = None
    swagger_favicon_url: Annotated[
        Optional[str],
        Doc(
            """
            String default URL where the Swagger favicon is located
            and used within OpenAPI documentation,

            Example: `https://esmerald.dev/statics/images/favicon.ico`.
            """
        ),
    ] = None
    with_google_fonts: Annotated[
        bool,
        Doc(
            """
            Boolean flag indicating if the google fonts shall be used
            in the ReDoc OpenAPI Documentation.
            """
        ),
    ] = True
    stoplight_js_url: Annotated[
        Optional[str],
        Doc(
            """
            Boolean flag indicating if the google fonts shall be used
            in the ReDoc OpenAPI Documentation.
            """
        ),
    ] = None
    stoplight_css_url: Annotated[
        Optional[str],
        Doc(
            """
            String default URL where the Stoplight CSS is located
            and used within OpenAPI documentation,

            Example: `https://unpkg.com/@stoplight/elements/styles.min.css`.
            """
        ),
    ] = None
    stoplight_url: Annotated[
        Optional[str],
        Doc(
            """
            String default relative URL where the Stoplight documentation
            shall be accessed in the application.

            Example: `/docs/elements`.
            """
        ),
    ] = None
    stoplight_favicon_url: Annotated[
        Optional[str],
        Doc(
            """
            String default URL where the Stoplight favicon is located
            and used within OpenAPI documentation,

            Example: `https://esmerald.dev/statics/images/favicon.ico`.
            """
        ),
    ] = None
    rapidoc_url: Annotated[
        Optional[str],
        Doc(
            """
            String default relative URL where the Rapidoc documentation
            shall be accessed in the application.

            Example: `/docs/rapidoc`.
            """
        ),
    ] = None
    rapidoc_js_url: Annotated[
        Optional[str],
        Doc(
            """
            String default URL where the Stoplight Javascript is located
            and used within OpenAPI documentation,

            This is used as the default if no [OpenAPIConfig](https://esmerald.dev/configurations/openapi/config/) is provided.
            """
        ),
    ] = None
    rapidoc_favicon_url: Annotated[
        Optional[str],
        Doc(
            """
            String default URL where the RapiDoc favicon is located
            and used within OpenAPI documentation,

            Example: `https://esmerald.dev/statics/images/favicon.ico`.
            """
        ),
    ] = None
    webhooks: Annotated[
        Optional[Sequence[Any]],
        Doc(
            """
            This is the same principle of the `routes` but for OpenAPI webhooks.

            Read more [about webhooks](https://esmerald.dev/routing/webhooks).

            When a webhook is added, it will automatically add them into the
            OpenAPI documentation.
            """
        ),
    ] = None

    def openapi(self, app: Any) -> Dict[str, Any]:
        """Loads the OpenAPI routing schema"""
        openapi_schema = get_openapi(
            app=app,
            title=self.title,
            version=self.version,
            openapi_version=self.openapi_version,
            summary=self.summary,
            description=self.description,
            routes=app.routes,
            tags=self.tags,
            servers=self.servers,
            terms_of_service=self.terms_of_service,
            contact=self.contact,
            license=self.license,
            webhooks=self.webhooks,
        )
        app.openapi_schema = openapi_schema
        return app.openapi_schema

    def enable(self, app: Any) -> None:
        """Enables the OpenAPI documentation"""
        if self.openapi_url:
            urls = {server.get("url") for server in self.servers}
            server_urls = set(urls)

            @get(path=self.openapi_url)  # type: ignore
            async def _openapi(request: Request) -> JSONResponse:
                root_path = request.scope.get("root_path", "").rstrip("/")

                if root_path not in server_urls:
                    if root_path and self.root_path_in_servers:
                        self.servers.insert(0, {"url": root_path})
                        server_urls.add(root_path)
                return JSONResponse(self.openapi(app))

            app.add_route(
                path="/",
                handler=_openapi,
                include_in_schema=False,
                activate_openapi=False,
            )

        if self.openapi_url and self.docs_url:

            @get(path=self.docs_url)  # type: ignore
            async def swagger_ui_html(
                request: Request,
            ) -> HTMLResponse:  # pragma: no cover
                root_path = request.scope.get("root_path", "").rstrip("/")
                openapi_url = root_path + self.openapi_url
                oauth2_redirect_url = self.swagger_ui_oauth2_redirect_url
                if oauth2_redirect_url:
                    oauth2_redirect_url = root_path + oauth2_redirect_url
                return get_swagger_ui_html(
                    openapi_url=openapi_url,
                    title=self.title + " - Swagger UI",
                    oauth2_redirect_url=oauth2_redirect_url,
                    init_oauth=self.swagger_ui_init_oauth,
                    swagger_ui_parameters=self.swagger_ui_parameters,
                    swagger_js_url=self.swagger_js_url,
                    swagger_favicon_url=self.swagger_favicon_url,
                    swagger_css_url=self.swagger_css_url,
                )

            app.add_route(
                path="/",
                handler=swagger_ui_html,
                include_in_schema=False,
                activate_openapi=False,
            )

        if self.swagger_ui_oauth2_redirect_url:

            @get(self.swagger_ui_oauth2_redirect_url)  # type: ignore
            async def swagger_ui_redirect(
                request: Request,
            ) -> HTMLResponse:  # pragma: no cover
                return get_swagger_ui_oauth2_redirect_html()

            app.add_route(
                path="/",
                handler=swagger_ui_redirect,
                include_in_schema=False,
                activate_openapi=False,
            )

        if self.openapi_url and self.redoc_url:

            @get(self.redoc_url)  # type: ignore
            async def redoc_html(request: Request) -> HTMLResponse:  # pragma: no cover
                root_path = request.scope.get("root_path", "").rstrip("/")
                openapi_url = root_path + self.openapi_url
                return get_redoc_html(
                    openapi_url=openapi_url,
                    title=self.title + " - ReDoc",
                    redoc_js_url=self.redoc_js_url,
                    redoc_favicon_url=self.redoc_favicon_url,
                    with_google_fonts=self.with_google_fonts,
                )

            app.add_route(
                path="/",
                handler=redoc_html,
                include_in_schema=False,
                activate_openapi=False,
            )

        if self.openapi_url and self.stoplight_url:

            @get(self.stoplight_url)  # type: ignore
            async def stoplight_html(
                request: Request,
            ) -> HTMLResponse:  # pragma: no cover
                root_path = request.scope.get("root_path", "").rstrip("/")
                openapi_url = root_path + self.openapi_url
                return get_stoplight_html(
                    openapi_url=openapi_url,
                    title=self.title + " - Stoplight Elements",
                    stoplight_js=self.stoplight_js_url,
                    stoplight_css=self.stoplight_css_url,
                    stoplight_favicon_url=self.stoplight_favicon_url,
                )

            app.add_route(
                path="/",
                handler=stoplight_html,
                include_in_schema=False,
                activate_openapi=False,
            )

        if self.openapi_url and self.rapidoc_url:

            @get(self.rapidoc_url)  # type: ignore
            async def rapidoc_html(
                request: Request,
            ) -> HTMLResponse:  # pragma: no cover
                root_path = request.scope.get("root_path", "").rstrip("/")
                openapi_url = root_path + self.openapi_url

                return get_rapidoc_ui_html(
                    openapi_url=openapi_url,
                    title=self.title + " - RapiDoc",
                    rapidoc_js_url=self.rapidoc_js_url,
                    rapidoc_favicon_url=self.rapidoc_favicon_url,
                )

            app.add_route(
                path="/",
                handler=rapidoc_html,
                include_in_schema=False,
                activate_openapi=False,
            )

        app.router.activate()



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from lilya._internal._path import clean_path
from lilya.staticfiles import StaticFiles
from lilya.types import ASGIApp
from pydantic import BaseModel, DirectoryPath, constr, field_validator
from typing_extensions import Annotated, Doc


class StaticFilesConfig(BaseModel):
    """
    An instance of [StaticFilesConfig](https://esmerald.dev/configurations/staticfiles/).

    This configuration is used to enable and serve static files via
    Esmerald application.

    **Example**

    ```python
    from esmerald import Esmerald
    from esmerald.config import StaticFilesConfig

    static_files_config = StaticFilesConfig(
        path="/static", directory=Path("static")
    )

    app = Esmerald(static_files_config=static_files_config)
    ```
    """

    path: Annotated[  # type: ignore
        constr(min_length=1),
        Doc(
            """
            The path for the statics.
            """
        ),
    ]
    directory: Annotated[
        Optional[Union[DirectoryPath, str, Path, Any]],
        Doc(
            """
            The directory for the statics in the format of a path like.

            Example: `/static`.
            """
        ),
    ] = None
    html: Annotated[
        bool,
        Doc(
            """
            Run in HTML mode. Automatically loads index.html for directories if such file exist.
            """
        ),
    ] = False
    packages: Annotated[
        Optional[List[Union[str, Tuple[str, str]]]],
        Doc(
            """
            A list of strings or list of tuples of strings of python packages.
            """
        ),
    ] = None
    check_dir: Annotated[
        bool,
        Doc(
            """
            Ensure that the directory exists upon instantiation.
            """
        ),
    ] = True

    @field_validator("path")
    def validate_path(cls, value: str) -> str:
        if "{" in value:
            raise ValueError("path parameters are not supported for static files")
        return clean_path(value)

    def _build_kwargs(
        self,
    ) -> Dict[str, Union[bool, int, DirectoryPath, List[Union[str, Tuple[str, str]]]]]:
        """
        Builds the necessary kwargs to create an StaticFiles object.
        """
        kwargs = {"html": self.html, "check_dir": self.check_dir}
        if self.packages:
            kwargs.update({"packages": self.packages})  # type: ignore
        if self.directory:
            kwargs.update({"directory": str(self.directory)})  # type: ignore
        return kwargs  # type: ignore

    def to_app(self) -> ASGIApp:
        """
        It can be three scenarios
        """

        return StaticFiles(**self._build_kwargs())  # type: ignore



from mimetypes import guess_type
from pathlib import PurePath
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from lilya.types import Receive, Scope, Send

from esmerald.enums import MediaType
from esmerald.responses.base import Response

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.background import BackgroundTask, BackgroundTasks
    from esmerald.protocols.template import TemplateEngineProtocol
    from esmerald.types import ResponseCookies


class TemplateResponse(Response):
    def __init__(
        self,
        template_name: str,
        template_engine: "TemplateEngineProtocol",
        status_code: int = 200,
        context: Optional[Dict[str, Any]] = None,
        background: Optional[Union["BackgroundTask", "BackgroundTasks"]] = None,
        headers: Optional[Dict[str, Any]] = None,
        cookies: Optional["ResponseCookies"] = None,
        media_type: Union[MediaType, str] = MediaType.HTML,
    ):
        if media_type == MediaType.JSON:  # we assume this is the default
            suffixes = PurePath(template_name).suffixes
            for suffix in suffixes:
                _type = guess_type("name" + suffix)[0]
                if _type:
                    media_type = _type
                    break
            else:
                media_type = MediaType.TEXT  # pragma: no cover

        self.template = template_engine.get_template(template_name)
        self.context = context or {}
        content = self.template.render(**context)
        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers,
            media_type=media_type,
            background=background,
            cookies=cookies,
        )

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:  # pragma: no cover
        request = self.context.get("request", {})
        extensions = request.get("extensions", {})
        if "http.response.template" in extensions:
            await send(
                {
                    "type": "http.response.template",
                    "template": self.template,
                    "context": self.context,
                }
            )
        await super().__call__(scope, receive, send)



from typing import Any, Dict, cast

from lilya.responses import JSONResponse

from esmerald.encoders import json_encoder


class BaseJSONResponse(JSONResponse):
    """
    Making sure it parses all the values from pydantic into dictionary.
    """

    @staticmethod
    def transform(value: Any) -> Dict[str, Any]:  # pragma: no cover
        """
        Makes sure that every value is checked and if it's a pydantic model then parses into
        a dict().
        """
        return cast(Dict[str, Any], json_encoder(value))


__all__ = ["BaseJSONResponse"]



from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Set,
    Tuple,
    Type,
    Union,
    cast,
    get_args,
    get_origin,
)

from lilya.datastructures import URL
from pydantic.fields import FieldInfo

from esmerald.enums import ParamType, ScopeType
from esmerald.exceptions import ImproperlyConfigured, ValidationErrorException
from esmerald.params import Cookie, Header, Path, Query
from esmerald.parsers import ArbitraryExtraBaseModel, HashableBaseModel
from esmerald.requests import Request
from esmerald.typing import Undefined
from esmerald.utils.constants import REQUIRED
from esmerald.utils.helpers import is_class_and_subclass, is_union
from esmerald.utils.schema import should_skip_json_schema

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.injector import Inject
    from esmerald.transformers.signature import Parameter, SignatureModel
    from esmerald.types import ConnectionType


class ParamSetting(NamedTuple):
    default_value: Any
    field_alias: str
    field_name: str
    is_required: bool
    param_type: ParamType
    field_info: FieldInfo


class Dependency(HashableBaseModel, ArbitraryExtraBaseModel):
    def __init__(
        self, key: str, inject: "Inject", dependencies: List["Dependency"], **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        self.key = key
        self.inject = inject
        self.dependencies = dependencies


def _merge_difference_parameters(difference: Set[ParamSetting]) -> Set[ParamSetting]:
    """
    Merge difference parameters based on field alias and requirement.

    Args:
        difference (Set[ParamSetting]): Set of difference parameters.

    Returns:
        Set[ParamSetting]: Merged difference parameters.
    """
    merged_result = set()

    for parameter in difference:
        if parameter.is_required or not any(
            param.field_alias == parameter.field_alias and param.is_required
            for param in difference
        ):
            merged_result.add(parameter)

    return merged_result


def merge_sets(first_set: Set[ParamSetting], second_set: Set[ParamSetting]) -> Set[ParamSetting]:
    """
    Merge two sets of parameter settings.

    Args:
        first_set (Set[ParamSetting]): First set of parameter settings.
        second_set (Set[ParamSetting]): Second set of parameter settings.

    Returns:
        Set[ParamSetting]: Merged set of parameter settings.
    """
    merged_result = first_set.intersection(second_set)
    difference = first_set.symmetric_difference(second_set)

    merged_result.update(_merge_difference_parameters(difference))

    return merged_result


def _get_default_value(field_info: FieldInfo) -> Any:
    """
    Get the default value from field information.

    Args:
        field_info (FieldInfo): Information about the field.

    Returns:
        Any: Default value of the field.
    """
    return field_info.default if field_info.default is not Undefined else None


def create_parameter_setting(
    allow_none: bool,
    field_info: FieldInfo,
    field_name: str,
    path_parameters: Set[str],
) -> ParamSetting:
    """
    Create a setting definition for a parameter.

    Args:
        allow_none (bool): Flag indicating if None is allowed.
        field_info (FieldInfo): Information about the field.
        field_name (str): Name of the field.
        path_parameters (Set[str]): Set of path parameters.

    Returns:
        ParamSetting: Parameter setting definition.
    """
    extra = cast("Dict[str, Any]", field_info.json_schema_extra) or {}
    is_required = extra.get(REQUIRED, True)
    default_value = _get_default_value(field_info)

    field_alias = extra.get(ParamType.QUERY) or field_name
    param_type = getattr(field_info, "in_", ParamType.QUERY)
    param: Union[Path, Header, Cookie, Query]

    if field_name in path_parameters:
        field_alias = field_name
        param_type = ParamType.PATH
        param = Path(default=default_value)
    elif extra.get(ParamType.HEADER):
        field_alias = extra[ParamType.HEADER]
        param_type = ParamType.HEADER
        param = Header(default=default_value)
    elif extra.get(ParamType.COOKIE):
        field_alias = extra[ParamType.COOKIE]
        param_type = ParamType.COOKIE
        param = Cookie(default=default_value)
    else:
        # Checking if the value should go to body or query params
        param = Query(default=default_value)

    if not field_info.alias:
        field_info.alias = field_name

    for key, _ in param._attributes_set.items():
        setattr(param, key, getattr(field_info, key, None))

    param_settings = ParamSetting(
        param_type=param_type,
        field_alias=field_alias,
        default_value=default_value,
        field_name=field_name,
        field_info=param,
        is_required=is_required and (default_value is None and not allow_none),
    )
    return param_settings


def _get_missing_required_params(params: Any, expected: Set[ParamSetting]) -> List[str]:
    """
    Get missing required parameters.

    Args:
        params (Any): Request parameters.
        expected (Set[ParamSetting]): Set of expected parameters.

    Returns:
        List[str]: List of missing required parameters.
    """
    missing_params = []
    for param in expected:
        if param.is_required and param.field_alias not in params:
            missing_params.append(param.field_alias)
    return missing_params


def get_request_params(
    params: Mapping[Union[int, str], Any], expected: Set[ParamSetting], url: URL
) -> Any:
    """
    Gather the parameters from the request.

    Args:
        params (Any): Request parameters.
        expected (Set[ParamSetting]): Set of expected parameters.
        url (URL): The URL.

    Returns:
        Any: The gathered parameters.

    Raises:
        ValidationErrorException: If required parameters are missing.
    """
    missing_params = _get_missing_required_params(params, expected)
    if missing_params:
        raise ValidationErrorException(
            f"Missing required parameter(s) {', '.join(missing_params)} for URL {url}."
        )

    values: Dict[Any, Any] = {}
    for param in expected:
        if not is_union(param.field_info.annotation):
            annotation = get_origin(param.field_info.annotation)
            origin = annotation or param.field_info.annotation

            if is_class_and_subclass(origin, (list, tuple)):
                values[param.field_name] = params.values()
            elif is_class_and_subclass(origin, dict):
                values[param.field_name] = dict(params.items()) if params else None
            else:
                values[param.field_name] = params.get(param.field_alias, param.default_value)
        elif is_union(param.field_info.annotation):
            arguments = get_args(param.field_info.annotation)
            if any(is_class_and_subclass(origin, (list, tuple)) for origin in arguments):
                values[param.field_name] = params.values()
            elif any(is_class_and_subclass(origin, dict) for origin in arguments):
                values[param.field_name] = dict(params.items()) if params else None
            else:
                values[param.field_name] = params.get(param.field_alias, param.default_value)
    return values


def get_connection_info(connection: "ConnectionType") -> Tuple[str, "URL"]:
    """
    Extacts the information from the ConnectionType.
    """
    method = connection.method if isinstance(connection, Request) else ScopeType.WEBSOCKET
    return method, connection.url


def get_signature(value: Any) -> Type["SignatureModel"]:
    try:
        return cast("Type[SignatureModel]", value.signature_model)
    except AttributeError as exc:
        raise ImproperlyConfigured(f"The 'signature' attribute for {value} is not set.") from exc


def get_field_definition_from_param(param: "Parameter") -> Tuple[Any, Any]:
    """
    This method will make sure that __future__ references are resolved by
    the Any type. This is necessary because the signature model will be
    generated before the actual type is resolved.
    """
    if param.optional:
        annotation = should_skip_json_schema(param)
    annotation = Any if isinstance(param.annotation, str) else param.annotation

    if param.default_defined:
        definition = annotation, param.default
    elif not param.optional:
        definition = annotation, ...
    else:
        definition = annotation, None
    return definition



from typing import Dict, Optional, Union

from pydantic import BaseModel, ConfigDict

from .callback import Callback
from .example import Example
from .header import Header
from .link import Link
from .parameter import Parameter
from .path_item import PathItem
from .reference import Reference
from .request_body import RequestBody
from .response import Response
from .schema import Schema
from .security_scheme import SecurityScheme


class Components(BaseModel):
    """Holds a set of reusable objects for different aspects of the OAS.

    All objects defined within the components object will have no effect
    on the API unless they are explicitly referenced from properties
    outside the components object.
    """

    model_config = ConfigDict(
        extra="ignore",
        json_schema_extra={
            "examples": [
                {
                    "schemas": {
                        "GeneralError": {
                            "type": "object",
                            "properties": {
                                "code": {"type": "integer", "format": "int32"},
                                "message": {"type": "string"},
                            },
                        },
                        "Category": {
                            "type": "object",
                            "properties": {
                                "id": {"type": "integer", "format": "int64"},
                                "name": {"type": "string"},
                            },
                        },
                        "Tag": {
                            "type": "object",
                            "properties": {
                                "id": {"type": "integer", "format": "int64"},
                                "name": {"type": "string"},
                            },
                        },
                    },
                    "parameters": {
                        "skipParam": {
                            "name": "skip",
                            "in": "query",
                            "description": "number of items to skip",
                            "required": True,
                            "schema": {"type": "integer", "format": "int32"},
                        },
                        "limitParam": {
                            "name": "limit",
                            "in": "query",
                            "description": "max records to return",
                            "required": True,
                            "schema": {"type": "integer", "format": "int32"},
                        },
                    },
                    "responses": {
                        "NotFound": {"description": "Entity not found."},
                        "IllegalInput": {"description": "Illegal input for operation."},
                        "GeneralError": {
                            "description": "General Error",
                            "content": {
                                "application/json": {
                                    "schema": {"$ref": "#/components/schemas/GeneralError"}
                                }
                            },
                        },
                    },
                    "securitySchemes": {
                        "api_key": {"type": "apiKey", "name": "api_key", "in": "header"},
                        "petstore_auth": {
                            "type": "oauth2",
                            "flows": {
                                "implicit": {
                                    "authorizationUrl": "http://example.org/api/oauth/dialog",
                                    "scopes": {
                                        "write:pets": "modify pets in your account",
                                        "read:pets": "read your pets",
                                    },
                                }
                            },
                        },
                    },
                }
            ]
        },
    )

    schemas: Optional[Dict[str, Schema]] = None
    """An object to hold reusable [Schema Objects](https://spec.openapis.org/oas/v3.1.0#schemaObject)."""

    responses: Optional[Dict[str, Union[Response, Reference]]] = None
    """An object to hold reusable [Response Objects](https://spec.openapis.org/oas/v3.1.0#responseObject)."""

    parameters: Optional[Dict[str, Union[Parameter, Reference]]] = None
    """An object to hold reusable [Parameter Objects](https://spec.openapis.org/oas/v3.1.0#parameterObject)."""

    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    """An object to hold reusable [Example Objects](https://spec.openapis.org/oas/v3.1.0#exampleObject)."""

    requestBodies: Optional[Dict[str, Union[RequestBody, Reference]]] = None
    """An object to hold reusable [Request Body Objects](https://spec.openapis.org/oas/v3.1.0#requestBodyObject)."""

    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    """An object to hold reusable [Header Objects](https://spec.openapis.org/oas/v3.1.0#headerObject)."""

    securitySchemes: Optional[Dict[str, Union[SecurityScheme, Reference]]] = None
    """An object to hold reusable [Security Scheme Objects](https://spec.openapis.org/oas/v3.1.0#securitySchemeObject)."""

    links: Optional[Dict[str, Union[Link, Reference]]] = None
    """An object to hold reusable [Link Objects](https://spec.openapis.org/oas/v3.1.0#linkObject)."""

    callbacks: Optional[Dict[str, Union[Callback, Reference]]] = None
    """An object to hold reusable [Callback Objects](https://spec.openapis.org/oas/v3.1.0#callbackObject)."""

    pathItems: Optional[Dict[str, Union[PathItem, Reference]]] = None
    """An object to hold reusable [Path Item Object](https://spec.openapis.org/oas/v3.1.0#pathItemObject)."""



from typing import Any, Callable, List, Optional

from lilya.background import Task, Tasks
from typing_extensions import ParamSpec

P = ParamSpec("P")


class BackgroundTask(Task):
    """
    `BackgroundTask` as a single instance can be easily achieved.

    **Example**

    ```python
    from pydantic import BaseModel

    from esmerald import BackgroundTask, JSONResponse, post


    class UserIn(BaseModel):
        email: str
        password: str


    async def send_email_notification(message: str):
        '''
        Sends an email notification
        '''
        send_notification(message)


    @post(
        "/register",
        background=BackgroundTask(send_email_notification, message="Account created"),
    )
    async def create_user(data: UserIn) -> JSONResponse:
        JSONResponse({"message": "Created"})
    ```
    """

    def __init__(self, func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs) -> None:
        super().__init__(func, *args, **kwargs)


class BackgroundTasks(Tasks):
    """
    Alternatively, the `BackgroundTasks` can also be used to be passed
    in.

    **Example**

    ```python
    from datetime import datetime

    from pydantic import BaseModel

    from esmerald import BackgroundTask, BackgroundTasks, JSONResponse, post


    class UserIn(BaseModel):
        email: str
        password: str


    async def send_email_notification(message: str):
        '''
        Sends an email notification
        '''
        send_notification(message)


    def write_in_file():
        with open("log.txt", mode="w") as log:
            now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            content = f"Notification sent @ {now}"
            log.write(content)


    @post(
        "/register",
        background=BackgroundTasks(
            tasks=[
                BackgroundTask(send_email_notification, message="Account created"),
                BackgroundTask(write_in_file),
            ]
        ),
    )
    async def create_user(data: UserIn) -> JSONResponse:
        JSONResponse({"message": "Created"})
    ```
    """

    def __init__(self, tasks: Optional[List[BackgroundTask]] = None):
        super().__init__(tasks=tasks)



from .asyncexitstack import AsyncExitStackMiddleware
from .authentication import BaseAuthMiddleware
from .clickjacking import XFrameOptionsMiddleware
from .cors import CORSMiddleware
from .csrf import CSRFMiddleware
from .gzip import GZipMiddleware
from .https import HTTPSRedirectMiddleware
from .security import SecurityMiddleware
from .sessions import SessionMiddleware
from .settings_middleware import RequestSettingsMiddleware
from .trustedhost import TrustedHostMiddleware

__all__ = [
    "AsyncExitStackMiddleware",
    "BaseAuthMiddleware",
    "CORSMiddleware",
    "CSRFMiddleware",
    "GZipMiddleware",
    "HTTPSRedirectMiddleware",
    "SessionMiddleware",
    "RequestSettingsMiddleware",
    "TrustedHostMiddleware",
    "XFrameOptionsMiddleware",
    "SecurityMiddleware",
]



import os
import sys
from typing import TYPE_CHECKING, Union

import click

from esmerald.core.directives.env import DirectiveEnv
from esmerald.core.directives.exceptions import DirectiveError
from esmerald.core.terminal import OutputColour, Print, Terminal

if TYPE_CHECKING:
    pass

printer = Print()
terminal = Terminal()


@click.option(
    "-p",
    "--port",
    type=int,
    default=8000,
    help="Port to run the development server.",
    show_default=True,
)
@click.option(
    "-r",
    "--reload",
    type=bool,
    default=True,
    help="Reload server on file changes.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--host",
    type=str,
    default="localhost",
    help="Server host. Tipically localhost.",
    show_default=True,
)
@click.option(
    "--debug",
    type=bool,
    default=True,
    help="Start the application in debug mode.",
    show_default=True,
    is_flag=True,
)
@click.option(
    "--log-level",
    type=str,
    default="debug",
    help="What log level should uvicorn run.",
    show_default=True,
)
@click.option(
    "--lifespan",
    type=str,
    default="on",
    help="Enable lifespan events.",
    show_default=True,
)
@click.option(
    "--settings", type=str, help="Any custom settings to be initialised.", required=False
)
@click.command(name="runserver")
def runserver(
    env: DirectiveEnv,
    port: int,
    reload: bool,
    host: str,
    debug: bool,
    log_level: str,
    lifespan: str,
    settings: Union[str, None] = None,
) -> None:
    """Starts the Esmerald development server.

    The --app can be passed in the form of <module>.<submodule>:<app> or be set
    as environment variable ESMERALD_DEFAULT_APP.

    Alternatively, if none is passed, Esmerald will perform the application discovery.

    It is strongly advised not to run this command in any other environment but development.
    This was designed to facilitate the development environment and should not be used in production.

    How to run: `esmerald runserver`
    """
    if getattr(env, "app", None) is None:
        error = (
            "You cannot specify a custom directive without specifying the --app or setting "
            "ESMERALD_DEFAULT_APP environment variable."
        )
        printer.write_error(error)
        sys.exit(1)

    if settings is not None:
        os.environ.setdefault("ESMERALD_SETTINGS_MODULE", settings)

    try:
        import uvicorn
    except ImportError:
        raise DirectiveError(detail="Uvicorn needs to be installed to run Esmerald.") from None

    app = env.app
    message = terminal.write_info(
        f"Starting {app.settings.environment} server @ {host}",
        colour=OutputColour.BRIGHT_CYAN,
    )
    terminal.rule(message, align="center")

    if settings is not None:
        custom_message = f"'{os.environ.get('ESMERALD_SETTINGS_MODULE')}'"
        terminal.rule(custom_message, align="center")

    if debug:
        app.debug = debug

    uvicorn.run(
        app=env.path,
        port=port,
        host=host,
        reload=reload,
        lifespan=lifespan,  # type: ignore
        log_level=log_level,
    )



from typing import Optional, Set

from pydantic import BaseModel
from typing_extensions import Annotated, Doc, Literal

from esmerald.types import HTTPMethod


class CSRFConfig(BaseModel):
    """
    An instance of [CRSFConfig](https://esmerald.dev/configurations/csrf/).

    This configuration is passed to the [CSRFMiddleware](https://esmerald.dev/middleware/middleware/#csrfmiddleware) and enables the middleware.

    !!! Tip
        You can creatye your own `CRSFMiddleware` version and pass your own
        configurations. You don't need to use the built-in version although it
        is recommended to do it so.

    **Example**

    ```python
    from esmerald import Esmerald
    from esmerald.config import CSRFConfig

    csrf_config = CSRFConfig(secret="your-long-unique-secret")

    app = Esmerald(csrf_config=csrf_config)
    ```
    """

    secret: Annotated[
        str,
        Doc(
            """
            The string used for the encryption/decryption and used to create an HMAC to sign
            the CSRF token.

            !!! Tip
                It is advised to use the same secret as the one in the settings to make it consistent.
            """
        ),
    ]
    cookie_name: Annotated[
        str,
        Doc(
            """
            The name of the CSRF cookie.
            """
        ),
    ] = "csrftoken"
    cookie_path: Annotated[
        str,
        Doc(
            """
            Name path of the CSRF cookie.
            """
        ),
    ] = "/"
    header_name: Annotated[
        str,
        Doc(
            """
            The header expected that will be expected in each request.
            """
        ),
    ] = "X-CSRFToken"
    secure: Annotated[
        bool,
        Doc(
            """
            Boolean flag when enabled sets `Secure` on the cookie.
            """
        ),
    ] = False
    httponly: Annotated[
        bool,
        Doc(
            """
            Boolean flag when enabled sets the cookie to be `httpsOnly`.
            """
        ),
    ] = False
    samesite: Annotated[
        Literal["lax", "strict", "none"],
        Doc(
            """
            The value to set in the `SameSite` attribute of the cookie.
            """
        ),
    ] = "lax"
    domain: Annotated[
        Optional[str],
        Doc(
            """
            Specifies which hosts can receive the cookie.
            """
        ),
    ] = None
    safe_methods: Annotated[
        Set[HTTPMethod],
        Doc(
            """
            A set of allowed safe methods that can set the cookie.
            """
        ),
    ] = {"GET", "HEAD"}



from esmerald.routing.apis._metaclasses import ListAPIMeta, SimpleAPIMeta
from esmerald.routing.apis._mixins import MethodMixin
from esmerald.routing.apis.base import View


class SimpleAPIView(View, MethodMixin, metaclass=SimpleAPIMeta):
    """The Esmerald SimpleAPIView class.

    This class has the same available parameters as the parent,
    `View`.

    **Example**

    ```python
    from esmerald import SimpleAPIView, delete, get, patch, post, put


    class UserAPI(SimpleAPIView):
        @get()
        async def get(self) -> str:
            ...

        @post()
        async def post(self) -> str:
            ...

        @put()
        async def put(self) -> str:
            ...

        @patch()
        async def patch(self) -> str:
            ...

        @delete()
        async def delete(self) -> None:
            ...
    ```
    """


class ListView(View, MethodMixin, metaclass=ListAPIMeta):
    """
    Base API for views returning lists.
    """

    ...


class APIView(View):
    """The Esmerald APIView class.

    The parameters available are the ones provided by the parent,
    `View` class.

    **Example**

    ```python
    from esmerald.permissions import DenyAll, IsAuthenticated
    from esmerald.requests import Request
    from esmerald.responses import JSONResponse
    from esmerald.routing.apis.views import APIView
    from esmerald.routing.handlers import delete, get, post


    class UserAPIView(APIView):
        path = "/users"
        permissions = [IsAuthenticated]

        @get(path="/")
        async def all_users(self, request: Request) -> JSONResponse:
            # logic to get all users here
            users = ...

            return JSONResponse({"users": users})

        @get(path="/deny", permissions=[DenyAll], description="API description")
        async def all_usersa(self, request: Request) -> JSONResponse:
            ...

        @get(path="/allow")
        async def all_usersb(self, request: Request) -> JSONResponse:
            users = ...
            return JSONResponse({"Total Users": users.count()})

        @post(path="/create")
        async def create_user(self, request: Request) -> None:
            # logic to create a user goes here
            ...

        @delete(path="/delete/{user_id}")
        async def delete_user(self, request: Request, user_id: str) -> None:
            # logic to delete a user goes here
            ...
    ```
    """

    ...


class Controller(APIView):
    """The Esmerald Controller class.

    **This is a drop-in replacement for the `APIView` class.**

    The parameters available are the ones provided by the parent,
    `View` class.

    **Example**

    ```python
    from esmerald.permissions import DenyAll, IsAuthenticated
    from esmerald.requests import Request
    from esmerald.responses import JSONResponse
    from esmerald.routing.apis.views import Controller
    from esmerald.routing.handlers import delete, get, post


    class UserController(Controller):
        path = "/users"
        permissions = [IsAuthenticated]

        @get(path="/")
        async def all_users(self, request: Request) -> JSONResponse:
            # logic to get all users here
            users = ...

            return JSONResponse({"users": users})

        @get(path="/deny", permissions=[DenyAll], description="API description")
        async def all_usersa(self, request: Request) -> JSONResponse:
            ...

        @get(path="/allow")
        async def all_usersb(self, request: Request) -> JSONResponse:
            users = ...
            return JSONResponse({"Total Users": users.count()})

        @post(path="/create")
        async def create_user(self, request: Request) -> None:
            # logic to create a user goes here
            ...

        @delete(path="/delete/{user_id}")
        async def delete_user(self, request: Request, user_id: str) -> None:
            # logic to delete a user goes here
            ...
    ```
    """

    ...
